import shutil

import h5py
import numpy as np
import numpy.testing as npt
import pytest

import pynbody
from pynbody import units


def setup_module() :
    global snap,subfind
    snap = pynbody.load('testdata/Test_NOSN_NOZCOOL_L010N0128/data/snapshot_103/snap_103.hdf5')
    subfind = pynbody.load('testdata/Test_NOSN_NOZCOOL_L010N0128/data/subhalos_103/subhalo_103')

def teardown_module() :
    global snap,subfind
    del snap
    del subfind

def test_standard_arrays() :
    """Check that the data loading works"""

    for s in [snap, subfind] :
        s.dm['pos']
        s.gas['pos']
        s.star['pos']
        s['pos']
        s['mass']
    #Load a second time to check that family_arrays still work
        s.dm['pos']
        s['vel']
        s['iord']
        s.gas['rho']
       # s.gas['u']
        s.star['mass']


def _h5py_copy_with_key_rename(src,dest):
    shutil.copy(src,dest)
    f = h5py.File(dest, 'r+')
    for tp in f:
        if "Mass" in tp:
            tp.move("Mass","Masses")
    f.close()




def test_alt_names():
    _h5py_copy_with_key_rename('testdata/Test_NOSN_NOZCOOL_L010N0128/data/snapshot_103/snap_103.hdf5',
                'testdata/Test_NOSN_NOZCOOL_L010N0128/data/snapshot_103/snap_103_altnames.hdf5')

    snap_alt = pynbody.load('testdata/Test_NOSN_NOZCOOL_L010N0128/data/snapshot_103/snap_103_altnames.hdf5')
    assert 'mass' in snap_alt.loadable_keys()
    assert all(snap_alt['mass']==snap['mass'])

def test_issue_256() :
    assert 'pos' in snap.loadable_keys()
    assert 'pos' in snap.dm.loadable_keys()
    assert 'pos' in snap.gas.loadable_keys()
    assert 'He' not in snap.loadable_keys()
    assert 'He' not in snap.dm.loadable_keys()
    assert 'He' in snap.gas.loadable_keys()

def test_write():
    ar_name = 'test_array'
    snap[ar_name] = np.random.uniform(0,1,len(snap))
    snap[ar_name].write()
    snap2 = pynbody.load('testdata/Test_NOSN_NOZCOOL_L010N0128/data/snapshot_103/snap_103.hdf5')
    v = snap[ar_name]
    with pytest.warns(UserWarning, match="Unable to infer units from HDF attributes"):
        v2 = snap2[ar_name]
    npt.assert_allclose(v, v2)

def test_hi_derivation():
    HI_answer = [  6.96499870e-06,   6.68348046e-06,   1.13855074e-05,
         1.10936027e-05,   1.40641633e-05,   1.67324738e-05,
         2.26228929e-05,   1.64661638e-05,   2.79337124e-05,
         3.32789555e-05,   2.38397192e-05,   5.11526743e-04,
         1.86211183e-01,   4.58309086e-02,   9.98117529e-02,
         1.76779058e-02,   8.30149935e-02,   1.08688537e-02,
         1.44146419e-07,   7.95141614e-08,   8.00568016e-05,
         5.40560080e-08,   2.73720754e-01,   2.91772885e-02,
         7.37755701e-04,   3.93431603e-02,   3.52700543e-03,
         1.46685188e-07,   4.55900305e-08,   3.85495273e-03,
         1.75020358e-07,   1.27841671e-01,   1.01551435e-07,
         3.23647121e-08,   8.22351949e-06,   1.03758201e-05,
         1.13115067e-05,   2.57878344e-05,   2.74634221e-05,
         5.34312023e-05,   2.55750061e-01,   3.83638138e-04,
         7.96613219e-03,   2.57835498e-03,   5.89219887e-08]

    assert np.allclose(subfind.halos()[0].gas['HI'][::100],HI_answer)

def test_hdf_ordering():
    # HDF files do not intrinsically specify the order in which the particle types occur
    # Because some operations may require stability, pynbody now imposes order by the particle type
    # number
    assert snap._family_slice[pynbody.family.gas] == slice(0, 2076907, None)
    assert snap._family_slice[pynbody.family.dm] == slice(2076907, 4174059, None)
    assert snap._family_slice[pynbody.family.star] == slice(4174059, 4194304, None)


def test_mass_in_header():
    f = pynbody.load("testdata/snap_028_z000p000.0.hdf5")
    f.physical_units()
    f['mass'] # load all masses
    assert np.allclose(f.dm['mass'][0], 3981879.2046075417)

    f = pynbody.load("testdata/snap_028_z000p000.0.hdf5")
    f.physical_units()
    # don't load all masses, allow it to be loaded for DM only
    assert np.allclose(f.dm['mass'][0], 3981879.2046075417)

def test_gadgethdf_style_units():
    f = pynbody.load("testdata/Test_NOSN_NOZCOOL_L010N0128/data/snapshot_103/snap_103.hdf5")
    npt.assert_allclose(f.st['InitialMass'].units.in_units("1.989e43 g h^-1"), 1.0,
                        rtol=1e-3)

def test_arepo_style_units():
    with pytest.warns(UserWarning, match=r"Masses are either stored in the header or have.*"):
        f = pynbody.load("testdata/arepo/agora_100.hdf5")
    npt.assert_allclose(f.st['EMP_InitialStellarMass'].units.in_units("1.989e42 g"),
                        1.0, rtol=1e-3)
    # I strongly suspect that the units in this file are wrong -- the masses are
    # in h^-1 units, so presumably these initial stellar masses should also be in
    # h^-1 units. This is backed up by checking that, numerically,
    #    (f.st['EMP_InitialStellarMass']/f.st['mass']).min() == 1.0
    # On the other hand, pynbody should just reflect back that error
    # to the user, really; we can't get involved in compensating for bugs in other codes,
    # or all hell will break loose. So, we check for the 'wrong' units.

    npt.assert_allclose(f.st['AREPOEMP_Metallicity'].units.in_units(1.0),
                        1.0, rtol=1e-5)
    # the above is a special case of a dimensionless array

    with pytest.warns(UserWarning, match="Unable to infer units from HDF attributes"):
        assert f.st['EMP_BirthTemperature'].units == units.NoUnit()
    # here is a case where no unit information is recorded in the file (who knows why)



import numpy as np
import numpy.testing as npt

import pynbody

np.random.seed(1)

def make_fake_bar(npart=100000, max=1, min=-1, barlength=.8, barwidth=0.05, phi=0, fraction=0.2):

    x = np.random.sample(int(npart*fraction))*(max-min) + min
    y = np.random.sample(int(npart*fraction))*(max-min) + min

    xbar = np.random.sample(npart, )*(barlength/2+barlength/2) - barlength/2
    ybar = np.random.sample(npart)*(barwidth/2+barwidth/2) - barwidth/2

    x = np.concatenate([x,xbar])
    y = np.concatenate([y,ybar])

    good = np.where(x**2 + y**2 < 1)[0]

    s = pynbody.snapshot.new(len(good))
    s['x'] = x[good]
    s['y'] = y[good]
    s['pos'].units = 'kpc'
    s['mass'] = 1.0
    s['mass'].units = 'Msol'
    s['vel'] = 1.0
    s['vel'].units = 'km s^-1'
    s.rotate_z(phi)
    return s

def test_fourier_profile():
    bar = make_fake_bar(phi=45)

    p = pynbody.analysis.profile.Profile(bar, nbins=50)

    assert(np.all(p['fourier']['amp'][2,4:20] > 0.1))
    assert(np.allclose(np.abs(p['fourier']['phi'][2,4:20]/2), np.pi/4.0, rtol=0.05))


def test_potential_profile_fp64():
    f = pynbody.new(100)
    coords = np.random.normal(size=(100,3))
    del f['pos']
    del f['mass']
    f['pos'] = np.array(coords,dtype=np.float64)
    f['eps'] = np.ones(100,dtype=np.float64)
    f['mass'] = np.ones(100,dtype=np.float64)
    p = pynbody.analysis.profile.Profile(f, nbins=50)
    p['pot']


def test_potential_profile_fp32():
    f = pynbody.new(100)
    coords = np.random.normal(size=(100,3))
    del f['pos']
    del f['mass']
    f['pos'] = np.array(coords,dtype=np.float32)
    f['eps'] = np.ones(100,dtype=np.float32)
    f['mass'] = np.ones(100,dtype=np.float32)
    p = pynbody.analysis.profile.Profile(f, nbins=50)
    p['pot']


def test_unique_hash_generation():
    f1 = pynbody.load("testdata/g15784.lr.01024")
    p1 = pynbody.analysis.profile.Profile(f1, nbins=50)
    p2 = pynbody.analysis.profile.Profile(f1[:1000], nbins=50)

    hash1 = p1._generate_hash_filename_from_particles()
    hash2 = p2._generate_hash_filename_from_particles()

    assert(hash1 != hash2)
    assert(type(hash1) is str)
    assert(type(hash2) is str)


def test_write_profile():
    f1 = pynbody.load("testdata/g15784.lr.01024")

    p = pynbody.analysis.profile.Profile(f1[:1000], nbins=50)
    p['rbins'], p['density']

    # Write profile and read again
    p.write()
    read_profile = pynbody.analysis.profile.Profile(f1[:1000], load_from_file=True)

    npt.assert_allclose(read_profile.min, p.min)
    npt.assert_allclose(read_profile.max, p.max)
    npt.assert_allclose(read_profile.nbins, p.nbins)
    npt.assert_allclose(read_profile['rbins'], p['rbins'])
    npt.assert_allclose(read_profile['density'], p['density'])



import pytest

import pynbody
from pynbody.halo.hop import HOPCatalogue


# Note: we do not use a module-wide fixture here to prevent caching of units
@pytest.fixture
def f():
    yield pynbody.load("testdata/output_00080")


@pytest.fixture
def halos(f):
    yield HOPCatalogue(f)

def test_get_halo(halos):
    assert len(halos[0]) == 37927
    assert len(halos[1]) == 20870
    assert (halos[1].dm['iord'][::1000] == [950469, 178234, 570544, 602770, 983503, 185854, 189546, 183353,
                                           971918, 189890, 981389, 978381, 184919, 201405, 227622, 180288,
                                           194977, 195167, 187596, 192792, 590108]).all()
    assert len(halos) == 369



import os.path
import warnings

import numpy as np

from .. import array, config_parser, snapshot, units
from ..snapshot import gadgethdf
from . import Halo, HaloCatalogue


class SubFindHDFSubhaloCatalogue(HaloCatalogue) :
    """
    Gadget's SubFind HDF Subhalo catalogue.

    Initialized with the parent FOF group catalogue and created
    automatically when an fof group is created
    """

    def __init__(self, group_id, group_catalogue) :
        super().__init__(group_catalogue.base)

        self._group_id = group_id
        self._group_catalogue = group_catalogue
        self.__calc_len()

    def __calc_len(self):
        next_group_id = self._group_id + 1
        next_first_subhalo = self._group_catalogue.nsubhalos
        while next_group_id < len(self._group_catalogue._fof_group_first_subhalo)-1:
            if self._group_catalogue._fof_group_first_subhalo[self._group_id + 1]!=-1:
                next_first_subhalo = self._group_catalogue._fof_group_first_subhalo[next_group_id]
                break
            next_group_id+=1


        self._len = (next_first_subhalo - self._group_catalogue._fof_group_first_subhalo[self._group_id])

    def __len__(self):
        return self._len

    def _get_halo(self, halo_number):
        if self.base is None :
            raise RuntimeError("Parent SimSnap has been deleted")

        if halo_number > len(self)-1 :
            raise ValueError("FOF group %d does not have subhalo %d" % (self._group_id, halo_number))

        # need this to index the global offset and length arrays
        absolute_id = self._group_catalogue._fof_group_first_subhalo[self._group_id] + halo_number

        # now form the particle IDs needed for this subhalo
        type_map = self.base._family_to_group_map

        halo_lengths = self._group_catalogue._subfind_halo_lengths
        halo_offsets = self._group_catalogue._subfind_halo_offsets

        # create the particle lists
        tot_len = 0
        for g_ptypes in list(type_map.values()) :
            for g_ptype in g_ptypes:
                tot_len += halo_lengths[g_ptype][absolute_id]

        plist = np.zeros(tot_len,dtype='int64')

        npart = 0
        for ptype in self.base._families_ordered():
            # family slice in the SubFindHDFSnap
            sl = self.base._family_slice[ptype]

            for g_ptype in type_map[ptype]:
                # add the particle indices to the particle list
                offset = halo_offsets[g_ptype][absolute_id]
                length = halo_lengths[g_ptype][absolute_id]
                ind = np.arange(sl.start + offset, sl.start + offset + length)
                plist[npart:npart+length] = ind
                npart += length

        return SubFindHDFSubHalo(halo_number, self._group_id, self._group_catalogue, self, self.base, plist)


    @property
    def base(self) :
        return self._base()
from .details import number_mapping, particle_indices
from .subhalo_catalogue import SubhaloCatalogue


class SubFindHDFHaloCatalogue(HaloCatalogue) :
    """Halo or group catalogue generated from subfind information stored inside a GadgetHDF snapshot
    """

    # Names of various groups and attributes in the hdf file (which vary in different versions of SubFind?)

    _fof_name = 'FOF'
    _header_name = 'FOF'
    _subfind_name = 'SUBFIND'
    _subfind_grnr_name = 'GrNr'
    _subfind_first_gr_name = 'SUBFIND/FirstSubOfHalo'

    _numgrps_name = 'Total_Number_of_groups'
    _numsubs_name = 'Total_Number_of_subgroups'

    _grp_offset_name = 'Offset'
    _grp_len_name = 'Length'

    _sub_offset_name = 'SUB_Offset'
    _sub_len_name = 'SUB_Length'

    def __init__(self, sim, subs=False, _inherit_data_from=None) :
        """Initialise a Subfind catalogue.

        By default, the FoF groups are imported, and subhalos are available via the 'subhalos' attribute of each
        halo object, e.g.

        >>> f = pynbody.load('path/to/snapshot.hdf5')
        >>> h = f.halos()
        >>> h[1].subhalos[2] # returns the third subhalo of FoF group 1

        *sim*: The SimSnap
        *subs*: If True, enumerate the subhalos instead of the groups. Otherwise, the jth subhalo of FOF group i
        is still available, as halo_cat[i].subhalos[j].
        *_inherit_data_from*: for internal use only, allows subhalo catalogue to share data with its parent FOF catalogue
        """

        self._sub_mode = subs
        self._hdf_files = self._get_catalogue_multifile(sim)

        super().__init__(sim, number_mapping.SimpleHaloNumberMapper(0, self.__get_length()))

        if _inherit_data_from:
            self.__inherit_data(_inherit_data_from)
        else:
            self.__init_halo_offset_data()
            self.__init_subhalo_relationships()
            self.__init_halo_properties()
            self.__reshape_multidimensional_properties()
            self.__reassign_properties_from_sub_to_fof()

        if not subs:
            self._subhalo_catalogue = type(self)(sim, subs=True, _inherit_data_from=self)

    def __inherit_data(self, parent):
        attrs_to_share = ["_fof_properties", "_sub_properties", "_fof_group_offsets", "_fof_group_lengths",
                         "_subfind_halo_offsets", "_subfind_halo_lengths"]
        for attr in attrs_to_share:
            setattr(self, attr, getattr(parent, attr))


    def _get_catalogue_multifile(self, sim):
        """Some variants of Subfind put all the particle data in the catalogue files, in which case the catalogue
        HDF files are already present in the base simulation. In other cases, notably Gadget4/Arepo, this must be
        overridden to locate the actual HDF5 files for the catalogue"""
        if not isinstance(sim, gadgethdf.SubFindHDFSnap):
            raise ValueError("SubFindHDFHaloCatalogue can only work with a SubFindHDFSnap simulation")

        return sim._hdf_files

    def __init_ignorable_keys(self):
        self.fof_ignore = list(map(str.strip,config_parser.get("SubfindHDF","FoF-ignore").split(",")))
        self.sub_ignore = list(map(str.strip,config_parser.get("SubfindHDF","Sub-ignore").split(",")))

        for t in list(self.base._family_to_group_map.values()):
            # Don't add SubFind particles ever as this list is actually spherical overdensity
            self.sub_ignore.append(t[0])
            self.fof_ignore.append(t[0])

    def __init_halo_properties(self):
        self.__init_ignorable_keys()
        self._fof_properties = self.__get_property_dictionary_from_hdf(self._fof_name)
        self._sub_properties = self.__get_property_dictionary_from_hdf(self._subfind_name)


    def __get_property_dictionary_from_hdf(self, hdf_key):

        props = self._get_properties_from_multifile(self._hdf_files, hdf_key)

        for property_key in list(props.keys()):
            arr_units = self._get_units(hdf_key, property_key)
            if property_key in props:
                props[property_key] = props[property_key].view(array.SimArray)
                props[property_key].units = arr_units
                props[property_key].sim = self.base

        return props

    def _get_properties_from_multifile(self, multifile, hdf_key):
        hdf0 = multifile.get_file0_root()
        props = {}
        for h in multifile.iterroot():
            for property_key in hdf0[hdf_key].keys():
                if property_key in self.fof_ignore or property_key not in h[hdf_key]:
                    continue
                if property_key in props:
                    props[property_key] = np.append(props[property_key], h[hdf_key][property_key][()])
                else:
                    props[property_key] = np.asarray(h[hdf_key][property_key])
        return props

    def _get_units(self, hdf_key, property_key):
        hdf0 = self._hdf_files.get_file0_root()
        return self.base._get_units_from_hdf_attr(hdf0[hdf_key][property_key].attrs)

    def __reshape_multidimensional_properties(self):
        self.__reshape_multidimensional_properties_one_dictionary(self._sub_properties, self._nsubhalos)
        self.__reshape_multidimensional_properties_one_dictionary(self._fof_properties, self._ngroups)

    def __reshape_multidimensional_properties_one_dictionary(self, properties_dict, expected_array_length):
        for key in list(properties_dict.keys()):
            # Test if there are no remainders, i.e. array is multiple of halo length
            # then solve for the case where this is 1, 2 or 3 dimension
            if len(properties_dict[key]) % expected_array_length == 0:
                ndim = len(properties_dict[key]) // expected_array_length
                if ndim > 1:
                    properties_dict[key] = properties_dict[key].reshape(expected_array_length, ndim)

    def __reassign_properties_from_sub_to_fof(self):
        reassign = []
        for k,v in self._sub_properties.items():
            if v.shape[0]==self._ngroups:
                reassign.append(k)

        for reassign_i in reassign:
            self._fof_properties[reassign_i] = self._sub_properties[reassign_i]
            del self._sub_properties[reassign_i]

    def __init_subhalo_relationships(self):
        nsub = 0
        nfof = 0
        self._subfind_halo_parent_groups = np.empty(self._nsubhalos, dtype=int)
        self._fof_group_first_subhalo = np.empty(self._ngroups, dtype=int)
        for h in self._hdf_files.iterroot():
            parent_groups = h[self._subfind_name].get(self._subfind_grnr_name, [])
            self._subfind_halo_parent_groups[nsub:nsub + len(parent_groups)] = parent_groups
            nsub += len(parent_groups)

            first_groups = h.get(self._subfind_first_gr_name,[])
            self._fof_group_first_subhalo[nfof:nfof + len(first_groups)] = first_groups
            nfof += len(first_groups)

    def __get_length(self):
        hdf0 = self._hdf_files.get_file0_root()
        if self._sub_mode:
            return int(hdf0[self._header_name].attrs[self._numsubs_name])
        else:
            return int(hdf0[self._header_name].attrs[self._numgrps_name])

    def __init_halo_offset_data(self):
        hdf0 = self._hdf_files.get_file0_root()

        self._fof_group_offsets = {}
        self._fof_group_lengths = {}
        self._subfind_halo_offsets = {}
        self._subfind_halo_lengths = {}

        self._ngroups = int(hdf0[self._header_name].attrs[self._numgrps_name])
        self._nsubhalos = int(hdf0[self._header_name].attrs[self._numsubs_name])

        for fam in self.base._families_ordered():
            ptypes = self.base._family_to_group_map[fam]
            for ptype in ptypes:
                self._fof_group_offsets[ptype] = np.empty(self._ngroups, dtype='int64')
                self._fof_group_lengths[ptype] = np.empty(self._ngroups, dtype='int64')
                self._subfind_halo_offsets[ptype] = np.empty(self._nsubhalos, dtype='int64')
                self._subfind_halo_lengths[ptype] = np.empty(self._nsubhalos, dtype='int64')

                curr_groups = 0
                curr_subhalos = 0

                for h in self._hdf_files:
                    # fof groups
                    offset = self._get_halodata_array_with_default(h, self._grp_offset_name, self._fof_name, ptype, [])
                    length = self._get_halodata_array_with_default(h, self._grp_len_name, self._fof_name, ptype, [])
                    self._fof_group_offsets[ptype][curr_groups:curr_groups + len(offset)] = offset
                    self._fof_group_lengths[ptype][curr_groups:curr_groups + len(offset)] = length
                    curr_groups += len(offset)

                    # subfind subhalos
                    offset = self._get_halodata_array_with_default(h, self._sub_offset_name, self._subfind_name, ptype, [])
                    length = self._get_halodata_array_with_default(h, self._sub_len_name, self._subfind_name, ptype, [])
                    self._subfind_halo_offsets[ptype][curr_subhalos:curr_subhalos + len(offset)] = offset
                    self._subfind_halo_lengths[ptype][curr_subhalos:curr_subhalos + len(offset)] = length
                    curr_subhalos += len(offset)
                if curr_groups!=self._ngroups:
                    warnings.warn(f"Incorrect number of groups recovered from HDF files. Expected {self._ngroups}, found {curr_groups}")
                    self._ngroups = curr_groups
                    self._fof_group_offsets[ptype] = self._fof_group_offsets[ptype][:curr_groups]
                    self._fof_group_lengths[ptype] = self._fof_group_lengths[ptype][:curr_groups]
                if curr_subhalos!=self._nsubhalos:
                    warnings.warn(f"Incorrect number of subhalos recovered from HDF files. Expected {self._nsubhalos}, found {curr_subhalos}")
                    self._nsubhalos = curr_subhalos
                    self._subfind_halo_offsets[ptype] = self._subfind_halo_offsets[ptype][:curr_groups]
                    self._subfind_halo_lengths[ptype] = self._subfind_halo_lengths[ptype][:curr_groups]

    def _get_halodata_array(self, hdf_file, array_name, halo_or_group, particle_type):
        # In gadget3 implementation, halo_or_group is not needed. In Gadget4 implementation (below), it is.
        return hdf_file[particle_type][array_name]

    def _get_halodata_array_with_default(self, hdf_file, array_name, halo_or_group, particle_type, default):
        try:
            return self._get_halodata_array(hdf_file, array_name, halo_or_group, particle_type)
        except KeyError:
            return default

    def _get_properties_one_halo(self, i):
        def extract(arr, i):
            if np.issubdtype(arr.dtype, np.integer):
                return arr[i]
            else:
                return units.get_item_with_unit(arr, i)

        properties = {}
        if self._sub_mode:
            for key in self._sub_properties:
                properties[key] = extract(self._sub_properties[key], i)
        else:
            for key in self._fof_properties:
                properties[key] = extract(self._fof_properties[key], i)
            properties['children'], = np.where(self._subfind_halo_parent_groups == i)
        return properties

    def _get_particle_indices_one_halo(self, number):
        if self.base is None :
            raise RuntimeError("Parent SimSnap has been deleted")

        if number > len(self)-1 :
            description = "Subhalo" if self._sub_mode else "Group"
            raise ValueError(f"{description} {number} does not exist")

        type_map = self.base._family_to_group_map

        if self._sub_mode:
            lengths = self._subfind_halo_lengths
            offsets = self._subfind_halo_offsets
        else:
            lengths = self._fof_group_lengths
            offsets = self._fof_group_offsets


        # create the particle lists
        tot_len = 0
        for g_ptypes in list(type_map.values()) :
            for g_ptype in g_ptypes:
                tot_len += lengths[g_ptype][number]

        plist = np.empty(tot_len,dtype='int64')

        self._write_pynbody_index_list_into_array(plist, offsets, lengths, number, type_map)

        return plist

    def _write_pynbody_index_list_into_array(self, particle_indices, halo_or_group_offsets,
                                                      halo_or_group_lengths, halo_or_group_index, type_map):
        npart = 0
        for ptype in self.base._families_ordered():
            # family slice in the SubFindHDFSnap
            sl = self.base._family_slice[ptype]

            for g_ptype in type_map[ptype]:
                # add the particle indices to the particle list
                offset = halo_or_group_offsets[g_ptype][halo_or_group_index]
                length = halo_or_group_lengths[g_ptype][halo_or_group_index]
                ind = np.arange(sl.start + offset, sl.start + offset + length)
                particle_indices[npart:npart + length] = ind
                npart += length
        return npart

    def _get_all_particle_indices(self):
        if self.base is None:
            raise RuntimeError("Parent SimSnap has been deleted")

        type_map = self.base._family_to_group_map

        if self._sub_mode:
            lengths = self._subfind_halo_lengths
            offsets = self._subfind_halo_offsets
        else:
            lengths = self._fof_group_lengths
            offsets = self._fof_group_offsets

        # create the particle lists
        tot_len = 0
        for g_ptypes in list(type_map.values()):
            for g_ptype in g_ptypes:
                tot_len += lengths[g_ptype].sum()

        plist = np.empty(tot_len, dtype='int64')
        boundaries = np.empty((len(self), 2), dtype='int64')

        location = 0
        for i in range(len(self)):
            npart = self._write_pynbody_index_list_into_array(plist[location:], offsets, lengths, i, type_map)
            boundaries[i] = (location, location := location+npart)

        assert location == tot_len

        return particle_indices.HaloParticleIndices(plist, boundaries)

    def _group_and_halo_from_halo_index(self, i):
        """Return the group, halo pair of indices from the overall halo index i

        Subfind stores a long list of halos, gathered together into groups. This maps from the index in the long
        list to the particular group and subhalo indices."""

        if i>=self._nsubhalos:
            raise ValueError("Subhalo out of range")

        group = np.argmax(self._fof_group_first_subhalo>i)-1
        halo = i - self._fof_group_first_subhalo[group]
        return group, halo

    def _get_subhalo_catalogue(self, parent_halo_number):
        if not self._sub_mode:
            props = self._get_properties_one_halo(parent_halo_number)
            return SubhaloCatalogue(self._subhalo_catalogue, props['children'])
        else:
            return SubhaloCatalogue(self._subhalo_catalogue, [])

    @classmethod
    def _can_load(cls, sim, subs=False):
        if isinstance(sim, gadgethdf.SubFindHDFSnap):
            return True
        else:
            return False

class SubFindFOFGroup(Halo) :
    """
    SubFind FOF group class
    """

    def __init__(self, group_id, *args) :
        """Construct a special halo representing subfind's FOF group"""
        super().__init__(group_id, *args)

        self._subhalo_catalogue = SubFindHDFSubhaloCatalogue(group_id, self._halo_catalogue)

        self._descriptor = "fof_group_"+str(group_id)

        self.properties.update(self._halo_catalogue.get_halo_properties(group_id, subs=False))


    def __getattr__(self, name):
        if name == 'sub':
            return self._subhalo_catalogue
        else :
            return super().__getattr__(name)

class SubFindHDFSubHalo(Halo) :
    """
    SubFind subhalo class
    """

    def __init__(self, halo_number, group_id, subfind_data_object, *args) :
        """Construct a special halo representing subfind's subhalo

        *halo_id*: The halo_id, where 0 is the first halo within the specified group
        *group_id*: The group across the entire snapshot
        *subfind_data_object*: The object which actually holds the HDF data

        Other arguments get passed to the standard halo constructor
        """
        super().__init__(halo_number, *args)
        self._group_id = group_id
        self._descriptor = "fof_group_%d_subhalo_%d"%(group_id, halo_number)

        # need this to index the global offset and length arrays
        absolute_id = subfind_data_object._fof_group_first_subhalo[self._group_id] + halo_number

        # load properties
        self.properties.update(subfind_data_object.get_halo_properties(absolute_id, subs=True))


class Gadget4SubfindHDFCatalogue(SubFindHDFHaloCatalogue):

    _fof_name = 'Group'
    _subfind_name = 'Subhalo'
    _header_name = 'Header'
    _subfind_grnr_name = 'SubhaloGroupNr'
    _subfind_first_gr_name = 'Group/GroupFirstSub'

    _numgrps_name = 'Ngroups_Total'
    _numsubs_name = 'Nsubhalos_Total'

    _grp_offset_name = 'GroupOffsetType'
    _grp_len_name = 'GroupLenType'

    _sub_offset_name = 'SubhaloOffsetType'
    _sub_len_name = 'SubhaloLenType'

    def __init__(self, sim, **kwargs):
        super().__init__(sim, **kwargs)
        i = 0
        for prog_or_desc in "prog", "desc":
            try:
                files = self._get_progenitor_or_descendant_multifile(sim, prog_or_desc)
            except FileNotFoundError:
                continue
            props = self._get_properties_from_multifile(files, 'Subhalo')
            self._sub_properties.update(props)

    def _get_halodata_array(self, hdf_file, array_name, halo_or_group, particle_type):
        return hdf_file[halo_or_group][array_name][:,int(particle_type[-1])]

    def _get_units(self, hdf_key, property_key):
        # Gadget4 doesn't seem to store unit information, so have a good guess
        if property_key == 'SubhaloVmax':
            dimensions = 'm s^-1'
        elif 'Mass' in property_key or '_M_' in property_key:
            dimensions = 'kg'
        elif 'Pos' in property_key or '_R_' in property_key or 'CM' in property_key or 'VmaxRad' in property_key:
            dimensions = 'm'
        elif 'Vel' in property_key:
            dimensions = 'm s^-1'
        else:
            return None

        return self.base.infer_original_units(dimensions)



    @classmethod
    def _get_catalogue_multifile(cls, sim):
        class Gadget4SubfindHdfMultiFileManager(gadgethdf.SubfindHdfMultiFileManager):
            _nfiles_groupname = cls._fof_name
            _nfiles_attrname = "NTask"
            _subgroup_name = None

        return Gadget4SubfindHdfMultiFileManager(cls._catalogue_filename(sim))

    @classmethod
    def _get_progenitor_or_descendant_multifile(cls, sim, prog_or_desc):
        class Gadget4SubfindHdfProgenitorsMultiFileManager(gadgethdf.SubfindHdfMultiFileManager):
            _nfiles_groupname = "Header"
            _nfiles_attrname = "NumFiles"
            _subgroup_name = None

        return Gadget4SubfindHdfProgenitorsMultiFileManager(cls._catalogue_filename(sim, "subhalo_"+prog_or_desc+"_"))

    @classmethod
    def _catalogue_filename(cls, sim, namestem ="fof_subhalo_tab_"):
        snapnum = os.path.basename(sim.filename).split("_")[-1]
        parent_dir = os.path.dirname(os.path.abspath(sim.filename))
        return os.path.join(parent_dir, namestem + snapnum)

    @classmethod
    def _can_load(cls, sim, **kwargs):
        file = cls._catalogue_filename(sim)
        if os.path.exists(file) and (file.endswith(".hdf5") or os.listdir(file)[0].endswith(".hdf5")) \
                or os.path.exists(file+".0.hdf5"):
            # very hard to figure out whether it's the right sort of hdf5 file without just going ahead and loading it
            try:
                cls(sim, **kwargs)
                return True
            except:
                return False


class ArepoSubfindHDFCatalogue(Gadget4SubfindHDFCatalogue):
    _numsubs_name = 'Nsubgroups_Total'

    _subfind_grnr_name = 'SubhaloGrNr'


    def _get_halodata_array(self, hdf_file, array_name, halo_or_group, particle_type):
        if array_name.endswith("OffsetType"):
            len_array_name = array_name.replace("OffsetType", "LenType")
            # this doesn't exist in arepo
            lens = super()._get_halodata_array(hdf_file, len_array_name, halo_or_group, particle_type)
            return np.concatenate(([0],np.cumsum(lens)[:-1]))
        else:
            return super()._get_halodata_array(hdf_file, array_name, halo_or_group, particle_type)

class TNGSubfindHDFCatalogue(ArepoSubfindHDFCatalogue):
    @classmethod
    def _catalogue_filename(cls, sim, namestem ="fof_subhalo_tab_"):
        snapnum = os.path.basename(sim.filename).split("_")[-1]
        parent_dir = os.path.dirname(os.path.dirname(os.path.abspath(sim.filename)))
        f = os.path.join(parent_dir, "groups_"+snapnum, namestem + snapnum)
        return f

    @classmethod
    def _get_catalogue_multifile(cls, sim):
        class TNGSubfindHdfMultiFileManager(gadgethdf.SubfindHdfMultiFileManager):
            _nfiles_groupname = "Header"
            _nfiles_attrname = "NumFiles"
            _subgroup_name = None

        return TNGSubfindHdfMultiFileManager(cls._catalogue_filename(sim))



import glob
import gzip
import os.path
import re
import warnings

import numpy as np

from .. import snapshot, util
from . import HaloCatalogue, HaloParticleIndices, logger
from .details.number_mapping import SimpleHaloNumberMapper, create_halo_number_mapper


class AHFCatalogue(HaloCatalogue):

    """
    Class to handle catalogues produced by Amiga Halo Finder (AHF).
    """

    def __init__(self, sim, make_grp=None, get_all_parts=None, use_iord=None, ahf_basename=None,
                 dosort=None, only_stat=None, write_fpos=True, halo_numbers='file-order',
                 ignore_missing_substructure=True,
                 **kwargs):
        """Initialize an AHFCatalogue.

        **kwargs** :

        *use_iord*: if True, the particle IDs in the Amiga catalogue
                    are taken to refer to the iord array. If False,
                    they are the particle offsets within the file. If
                    None, the parameter defaults to True for
                    GadgetSnap, False otherwise.

        *ahf_basename*: specify the basename of the AHF halo catalog
                        files - the code will append 'halos',
                        'particles', and 'substructure' to this
                        basename to load the catalog data.

        *halo_numbers*: specify how to number the halos. Options are:
                             'ahf' (default): use the halo numbers written in the AHF halos file,
                                              or a zero-based indexing if none is present
                             'file-order': zero-based indexing of the halos
                             'v1': one-based indexing of the halos, compatible with pynbody v1.x
                             'length-order': sort by the number of particles in each halo, with the
                                             halo with most particles being halo 0
                             'length-order-v1': as length-order, but indexing from halo 1,
                                                compatible with dosort=True in pynbody v1

        *ignore_missing_substructure*: if True (default), the code will not
                            raise an exception if the substructure file is
                            missing or corrupt. If False, it will raise an exception.

        Deprecated kwargs:

        *make_grp*: if True a 'grp' array is created in the underlying
                    snapshot specifying the lowest level halo that any
                    given particle belongs to. If it is False, no such
                    array is created; if None, the behaviour is
                    determined by the configuration system.

        *get_all_parts*: if True, the particle file is loaded for all halos.
                    Suggested to keep this None, as this is memory intensive.
                    The default function is to load in this data as needed.

        *dosort*: equivalent to halo_numbers='length-order'

        *only_stat*: specify that you only wish to collect the halo
                    properties stored in the AHF_halos file and not
                    worry about particle information [deprecated, use
                    get_dummy_halo instead]

        """


        if use_iord is None:
            use_iord = isinstance(
                sim.ancestor,
                (snapshot.gadget.GadgetSnap, snapshot.gadgethdf.GadgetHDFSnap)
            )

        self._use_iord = use_iord
        self._only_stat = only_stat
        self._try_writing_fpos = write_fpos

        if only_stat:
            warnings.warn(DeprecationWarning("only_stat keyword is deprecated; instead, use the catalogue's get_dummy_halo method"))

        if ahf_basename is not None:
            self._ahfBasename = ahf_basename
        else:
            self._ahfBasename = self._infer_ahf_basename(sim)

        self._determine_format_revision_from_filename()

        logger.info("AHFCatalogue loading halo properties")
        self._load_ahf_halo_properties(self._ahfBasename + 'halos')

        if dosort:
            warnings.warn(DeprecationWarning("dosort keyword is deprecated; instead pass halo_numbers='length-order-v1'"))
            halo_numbers = 'length-order-v1'

        number_mapper = self._setup_halo_numbering(halo_numbers)

        super().__init__(sim, number_mapper)

        self._remap_host_halo_property()

        if self._use_iord:
            self._init_iord_to_fpos()

        try:
            self._load_ahf_substructure(self._ahfBasename + 'substructure')
        except (KeyError, ValueError, FileNotFoundError):
            if not ignore_missing_substructure:
                raise
            logger.error("Unable to load AHF substructure file; continuing without. To expose the underlying problem as an exception, pass ignore_missing_substructure=False to the AHFCatalogue constructor")

        if make_grp:
            warnings.warn(DeprecationWarning("make_grp keyword is deprecated; instead, use the catalogue's get_group_array method"))
            self.base['grp'] = self.get_group_array()

        if get_all_parts:
            warnings.warn(
                DeprecationWarning("get_all_parts keyword is deprecated; instead, use the catalogue's load_all method"))
            self.load_all()

        logger.info("AHFCatalogue loaded")

    def _setup_halo_numbering(self, halo_numbers):
        has_id = 'ID' in self._halo_properties[0]
        if has_id:
            self._ahf_own_number_mapper = create_halo_number_mapper([properties['ID'] for properties in self._halo_properties])
        else:
            # if no explicit IDs, ahf implicitly numbers starting at 0 in file order
            self._ahf_own_number_mapper = SimpleHaloNumberMapper(0, len(self._halo_properties))

        if halo_numbers == 'v1':
            number_mapper = SimpleHaloNumberMapper(1, len(self._halo_properties))
        elif halo_numbers == 'file-order':
            number_mapper = SimpleHaloNumberMapper(0, len(self._halo_properties))
        elif halo_numbers == 'length-order' or halo_numbers == 'length-order-v1':
            npart = np.array([properties['npart'] for properties in self._halo_properties])
            osort = np.argsort(-npart)  # this is better than argsort(npart)[::-1], because it's stable
            if halo_numbers.endswith('v1'):
                number_mapper = create_halo_number_mapper(osort+1)
            else:
                number_mapper = create_halo_number_mapper(osort)
        elif halo_numbers == 'ahf':
            number_mapper = self._ahf_own_number_mapper
        else:
            raise ValueError(f"halo_numbers keyword {halo_numbers} not recognised")

        return number_mapper

    def _remap_host_halo_property(self):
        """When loaded from the .halos file, hostHalo is the AHF halo number of the host halo.
        This maps it onto whatever halo number pynbody is using."""

        if 'hostHalo' in self._halo_properties[0]:
            for properties in self._halo_properties:
                if properties['hostHalo'] != -1:
                    properties['hostHalo'] = self._number_mapper.index_to_number(
                        self._ahf_own_number_mapper.number_to_index(properties['hostHalo'])
                    )

    def _determine_format_revision_from_filename(self):
        if self._ahfBasename.split("z")[-2][-1] == ".":
            self._is_new_format = True
        else:
            self._is_new_format = False

    @classmethod
    def _infer_ahf_basename(cls, sim):
        candidates = cls._list_candidate_ahf_basenames(sim)
        if len(candidates) == 1:
            candidate = candidates[0]
        elif len(candidates) == 0:
            raise FileNotFoundError("No candidate AHF catalogue found; try specifying a catalogue using the "
                                    "ahf_basename keyword")
        else:
            candidate = candidates[0]
            warnings.warn(f"Multiple candidate AHF catalogues found; using {candidate}. To specify a different "
                          "catalogue, use the ahf_basename keyword, or move the other catalogues.")
        return util.cutgz(candidate)[:-9]


    def _write_fpos(self):
        try:
            np.savetxt(self._ahfBasename+'fpos', self._fpos, fmt="%d")
        except OSError:
            warnings.warn("Unable to write AHF_fpos file; performance will be reduced. Pass write_fpos=False to halo constructor to suppress this message, or use load_all() method to remove need for storing fpos information.")
        self._try_writing_fpos = False

    @property
    def base(self):
        return self._base()

    def _get_file_positions(self):
        """Get the starting positions of each halo's particle information within the
        AHF_particles file for faster access later"""
        if not hasattr(self, "_fpos"):
            if os.path.exists(self._ahfBasename + 'fpos'):
                self._fpos = np.loadtxt(self._ahfBasename+'fpos', dtype=int)
            else:
                self._fpos = np.empty(len(self._number_mapper), dtype=int)
                with util.open_(self._ahfBasename + 'particles') as f:
                    nhalo = int(f.readline().strip())
                    assert nhalo == len(self._number_mapper)
                    for hnum in range(nhalo):
                        npart = int(f.readline().strip())
                        assert npart == self._halo_properties[hnum]['npart']
                        self._fpos[hnum] = f.tell()
                        for i in range(npart):
                            f.readline()
                if self._try_writing_fpos:
                    if not os.path.exists(self._ahfBasename + 'fpos'):
                        self._write_fpos()

        return self._fpos

    def _load_ahf_particle_block(self, f, nparts):
        """Load the particles for the next halo described in particle file f"""
        if self._is_new_format:
            if not isinstance(f, gzip.GzipFile):
                data = np.fromfile(
                    f,
                    dtype=int,
                    sep=" ",
                    count=nparts * 2
                )[::2]
                data = np.ascontiguousarray(data)
            else:
                # unfortunately with gzipped files there does not
                # seem to be an efficient way to load nparts lines
                data = np.empty(nparts, dtype=int)
                for i in range(nparts):
                    data[i] = int(f.readline().split()[0])

            data = self._ahf_to_pynbody_particle_ids(data)
        else:
            if not isinstance(f, gzip.GzipFile):
                data = np.fromfile(f, dtype=int, sep=" ", count=nparts)
            else:
                # see comment above on gzipped files
                data = np.empty(nparts, dtype=int)
                for i in range(nparts):
                    data[i] = int(f.readline())
        data.sort()
        return data

    def _ahf_to_pynbody_particle_ids(self, data):
        ng = len(self.base.gas)
        nd = len(self.base.dark)
        ns = len(self.base.star)
        nds = nd + ns
        if self._use_iord:
            data = self._iord_to_fpos.map_ignoring_order(data)
        elif isinstance(self.base, snapshot.ramses.RamsesSnap):
            # AHF only expects three families, DM, star, gas in this order
            # and generates iords on disc according to this rule
            # For classical Ramses snapshots, this is perfectly adequate, but
            # for more modern outputs that have extra tracers, BHs families
            # we need to offset the ids to return the correct slicing
            # TODO These tests on snapshot type might not be necessary
            #  as using the family logic properly should be general.
            #  It is currently kept to ensure 100% backwards compatibility with previous behaviour,
            #  as this code is not explicitly checked by the pynbody test distribution

            if len(self.base) != nd + ns + ng:  # We have extra families to the base ones
                # First identify DM, star and gas particles in AHF
                ahf_dm_mask = data < nd
                ahf_star_mask = (data >= nd) & (data < nds)
                ahf_gas_mask = data >= nds

                # Then offset them by DM family start, to account for
                # additional families before it, e.g. gas tracers
                data[np.where(ahf_dm_mask)] += self.base._get_family_slice('dm').start

                # Star ids used to start at NDM, now they start with the star family slice
                offset = self.base._get_family_slice('star').start - nd
                data[np.where(ahf_star_mask)] += offset

                # Gas ids were greater than NDM + NSTAR, now they start with the gas slice
                offset = self.base._get_family_slice('gas').start - nds
                data[np.where(ahf_gas_mask)] += offset
        elif not isinstance(self.base, snapshot.nchilada.NchiladaSnap):
            hi_mask = data >= nds
            data[np.where(hi_mask)] -= nds
            data[np.where(~hi_mask)] += ng
        else:
            st_mask = (data >= nd) & (data < nds)
            g_mask = data >= nds
            data[np.where(st_mask)] += ng
            data[np.where(g_mask)] -= ns
        return data

    def _get_particle_indices_one_halo(self, halo_number):
        fpos = self._get_file_positions()
        file_index = self._number_mapper.number_to_index(halo_number)
        with util.open_(self._ahfBasename + 'particles') as f:
            f.seek(fpos[file_index],0)
            ids = self._load_ahf_particle_block(f, nparts=self._halo_properties[file_index]['npart'])
        return ids

    def _get_all_particle_indices(self):
        boundaries = np.cumsum([0] + [properties['npart'] for properties in self._halo_properties])
        boundaries = np.vstack((boundaries[:-1], boundaries[1:])).T
        particle_ids = np.empty(boundaries[-1,1], dtype=int)
        with util.open_(self._ahfBasename + 'particles') as f:
            f.readline()
            for properties, (start, end) in zip(self._halo_properties, boundaries):
                f.readline()
                particle_ids[start:end] = self._load_ahf_particle_block(f, nparts=properties['npart'])


        return HaloParticleIndices(particle_ids=particle_ids, boundaries=boundaries)


    def _get_properties_one_halo(self, i):
        return self._halo_properties[self._number_mapper.number_to_index(i)]

    def _load_ahf_halo_properties(self, filename):
        # Note: we need to open in 'rt' mode in case the AHF catalogue
        # is gzipped.
        with util.open_(filename, "rt") as f:
            first_line = f.readline()
            lines = f.readlines()

        self._halo_properties = [{} for h in range(len(lines))]

        # get all the property names from the first, commented line
        # remove (#)
        fields = first_line.replace("#", "").split()
        keys = [re.sub(r'\([0-9]*\)', '', field)
                for field in fields]

        if self._is_new_format:
            # fix for column 0 being a non-column in some versions of the AHF
            # output
            if keys[0] == '':
                keys = keys[1:]

        for properties, line in zip(self._halo_properties, lines):
            values = [
                float(x) if any(_ in x for _ in (".", "e", "nan", "inf"))
                else int(x)
                for x in line.split()
            ]
            # XXX Unit issues!  AHF uses distances in Mpc/h, possibly masses as
            # well
            if not self._is_new_format:
                values = values[1:]

            for key, value in zip(keys, values):
                properties[key] = value

    def _load_ahf_substructure(self, filename):
        with util.open_(filename) as f:
            lines = f.readlines()
        logger.info("AHFCatalogue loading substructure")

        if len(lines[0].split())!=2:
            # some variants of the format start with a line count, which we ignore
            lines = lines[1:]

        for halo_line, children_line in zip(lines[::2], lines[1::2]):
            haloid, _nsubhalos = (int(x) for x in halo_line.split())
            halo_index = self._ahf_own_number_mapper.number_to_index(haloid)
            halo_number = self._number_mapper.index_to_number(halo_index)
            children_ahf_numbering = [int(x) for x in children_line.split()]
            children_index = self._ahf_own_number_mapper.number_to_index(children_ahf_numbering)
            children_pynbody_numbering = self._number_mapper.index_to_number(children_index)

            self._halo_properties[halo_index]['children'] = children_pynbody_numbering
            for child_index in children_index:
                self._halo_properties[child_index]['parent'] = halo_number


    @staticmethod
    def _list_candidate_ahf_basenames(sim):
        candidates = set(glob.glob(f"{sim._filename}*z*particles*"))
        # use a set to ensure that no duplicates can be produced
        # This could arise in an edge case where _filename is a directory
        # and having the "/" at the end of it would lead to a first detection here
        # and a second one again below

        if os.path.isdir(sim._filename):
            candidates = candidates.union(glob.glob(os.path.join(sim._filename, "*z*particles*")))

        return list(candidates)

    @classmethod
    def _can_load(cls, sim, ahf_basename=None, **kwargs):
        if ahf_basename:
            return True
        candidates = cls._list_candidate_ahf_basenames(sim)
        number_ahf_file_candidates = len(candidates)
        return number_ahf_file_candidates > 0



import glob
import os

import numpy as np
import pytest

import pynbody

sink_filename = "testdata/ramses_new_format_partial_output_00001/sink_00001.csv"
sink_filename_moved = sink_filename+".temporarily_moved"


def setup_module():
    global f
    f = pynbody.load("testdata/ramses_new_format_partial_output_00001")

def test_family_lengths():
    assert len(f.dm)==136006
    assert len(f.star)==12236
    assert len(f.gas)==196232
    assert len(f.bh)==2
    assert pynbody.family.bh in f.families()

def test_properties():
    np.testing.assert_almost_equal(f.properties['a'], 1.0)
    np.testing.assert_almost_equal(f.properties['h'], 0.01)
    np.testing.assert_almost_equal(f.properties['omegaM0'], 1.0)

def test_sink_variables():
    np.testing.assert_allclose(f.bh['pos'], [[2.5e-2,2.5e-2,2.5e-2],
                                             [3.5e-2,4.5e-2,5.5e-2]])
    assert str(f.bh['pos'].units)=="3.09e+21 cm"
    assert (f.bh['id']==np.array([1,2])).all()


def _no_bh_family_present(f):
    return (len(f.bh)==0) and (pynbody.family.bh not in f.families())

@pytest.fixture
def backup_sinkfile():
    try:
        os.rename(sink_filename, sink_filename_moved)
        yield sink_filename
    finally:
        os.rename(sink_filename_moved, sink_filename)


def test_no_sink_file(backup_sinkfile):
    # No warning should be raised
    f_no_sink = pynbody.load("testdata/ramses_new_format_partial_output_00001")
    assert _no_bh_family_present(f_no_sink)

def test_garbled_sink_file(backup_sinkfile):
    with open(sink_filename, "w") as tfile:
        tfile.write("1,2,3\r\n")

    warn_str = r"Unexpected format in file .*\.csv -- sink data has not been loaded"
    with pytest.warns(UserWarning, match=warn_str):
        f_garbled_sink = pynbody.load("testdata/ramses_new_format_partial_output_00001")
    assert _no_bh_family_present(f_garbled_sink)

    with open(sink_filename,"w") as tfile:
        for i in range(4):
            tfile.write("1,2,3\r\n")

    with pytest.warns(UserWarning, match=warn_str):
        f_garbled_sink = pynbody.load("testdata/ramses_new_format_partial_output_00001")

    assert _no_bh_family_present(f_garbled_sink)

def test_load_pos():
    loaded_vals = f.dm['pos'][::5001]
    test_vals = [[ 0.01105415,  0.02300947,  0.00724412],
       [ 0.02122232,  0.02265619,  0.00677652],
       [ 0.02018343,  0.02035065,  0.0095223 ],
       [ 0.02269471,  0.02136712,  0.00620482],
       [ 0.02287763,  0.01224741,  0.01737309],
       [ 0.01815873,  0.01128699,  0.02247262],
       [ 0.02065698,  0.00701042,  0.02248414],
       [ 0.01930041,  0.01295866,  0.01411238],
       [ 0.02427911,  0.01443727,  0.02082101],
       [ 0.02015258,  0.01861216,  0.01329177],
       [ 0.02016469,  0.01742381,  0.02150717],
       [ 0.02356023,  0.02185429,  0.01414604],
       [ 0.01941696,  0.02358911,  0.01747907],
       [ 0.02231799,  0.02193878,  0.02197457],
       [ 0.02463229,  0.02258142,  0.02418932],
       [ 0.0129505 ,  0.01908118,  0.0244578 ],
       [ 0.0170727 ,  0.01979101,  0.02121195],
       [ 0.01921728,  0.02407719,  0.02357735],
       [ 0.01402912,  0.02266267,  0.01451735],
       [ 0.01251412,  0.01753267,  0.01765179],
       [ 0.01846307,  0.01644008,  0.02239697],
       [ 0.01017412,  0.01582988,  0.02493915],
       [ 0.00980422,  0.02452824,  0.02011278],
       [ 0.01242898,  0.02346281,  0.01821639],
       [ 0.01668166,  0.01431568,  0.01471138],
       [ 0.0067898 ,  0.01912026,  0.02174331],
       [ 0.02104224,  0.01082964,  0.02712232],
       [ 0.02348118,  0.00900178,  0.028894  ]]

    np.testing.assert_allclose(loaded_vals,test_vals, atol=1e-7)

def test_load_iord():
    loaded_vals = f.star['iord'][::1001]
    np.testing.assert_equal(loaded_vals,
                            [1001684, 1006393, 1039350, 1099620, 1035230, 1086365, 1000185,
                             1000429, 1077059, 1011855, 1096842, 1013422, 1020021])

def _make_virtual_output_with_no_ptcls():
    if os.path.exists("testdata/ramses_new_format_partial_no_ptcls_output_00001"):
        return

    os.mkdir("testdata/ramses_new_format_partial_no_ptcls_output_00001")
    for i in glob.glob("testdata/ramses_new_format_partial_output_00001/*"):
        if "part_00001" not in i and '.csv' not in i:
            os.symlink("../../"+i, i.replace("ramses_new_format_partial_output","ramses_new_format_partial_no_ptcls_output"))

def test_load_no_ptcls():
    _make_virtual_output_with_no_ptcls()
    f = pynbody.load("testdata/ramses_new_format_partial_no_ptcls_output_00001")
    assert len(f.dm)==0
    assert len(f.star)==0
    assert len(f.gas) == 196232

def test_loaded_namelist():
    assert f._namelist['cosmo'] is False
    assert f._namelist['hydro'] is True
    assert f._namelist['z_ave'] == 0.1
    assert f._namelist.get("non_existent_key") is None



import glob
import os
import time

import numpy as np
import pytest

import pynbody


def setup_module():
    global f, h
    f = pynbody.load("testdata/g15784.lr.01024")
    h = f.halos()


def teardown_module():
    global f, h
    del f, h


def test_center():
    global f, h
    with pynbody.analysis.halo.center(h[0]):
        np.testing.assert_allclose(
            f['pos'][0], [-0.0137471,  -0.00208458, -0.04392379], atol=1e-4)


def test_align():
    global f, h
    with pynbody.analysis.angmom.faceon(h[0]):
        np.testing.assert_allclose(f['pos'][:2], [[-0.010718, -0.001504, -0.044783],
                                                  [-0.010026,  0.002441, -0.04465 ]], atol=1e-5)

        np.testing.assert_allclose(f['vel'][:2], [[ 0.017203,  0.01848 , -0.019859],
                                                  [ 0.051333,  0.027357, -0.010303]], atol=1e-5)


def test_virialradius():
    global f, h
    with pynbody.analysis.halo.center(h[0]):
        start = time.time()
        vrad = pynbody.analysis.halo.virial_radius(f)
        print ("time=",time.time()-start)
        np.testing.assert_allclose(vrad, 0.005946911872, atol=1.e-5)


def test_ssc_bighalo():
    s = pynbody.load('testdata/Test_NOSN_NOZCOOL_L010N0128/data/subhalos_103/subhalo_103')
    s.physical_units()
    h = s.halos()
    pynbody.analysis.halo.center(h[1])
    assert h[1]['r'].min()<0.02


def test_binning_hmf():
    subfind = pynbody.load('testdata/Test_NOSN_NOZCOOL_L010N0128/data/subhalos_103/subhalo_103')

    h = subfind.halos()
    assert len(h) == 4226

    with pytest.warns(UserWarning, match=r"Halo finder masses not provided\. Calculating them.*"):
        center, means, err = pynbody.analysis.hmf.simulation_halo_mass_function(
            subfind,
            log_M_min=8,
            log_M_max=14,
            delta_log_M=0.5,
            subsample_catalogue=100
        )

    assert len(means) == len(center) == len(err) == 12

    np.testing.assert_allclose(center, [2.08113883e+08,   6.58113883e+08,   2.08113883e+09,   6.58113883e+09,
                                        2.08113883e+10,   6.58113883e+10,   2.08113883e+11,   6.58113883e+11,
                                        2.08113883e+12,   6.58113883e+12,   2.08113883e+13,   6.58113883e+13], atol=1e-5)

    np.testing.assert_allclose(means, [0., 0.05600012,  0.02200005,  0.00600001,  0, 0., 0.002,
                                        0., 0.,          0.,          0.,          0.], atol=1e-5)

    np.testing.assert_allclose(err, [0., 0.01058303,  0.00663326, 0.00346411,  0.,  0., 0.002,
                                     0.,          0.,         0.,          0.,          0.], atol=1e-5)



import weakref

import numpy as np

from . import snapshot


class Transformation:

    def __init__(self, f, defer=False):
        if isinstance(f, snapshot.SimSnap):
            self.sim = f
            self.next_transformation = None
        elif isinstance(f, Transformation):
            self.sim = None
            self.next_transformation = f
        else:
            raise ValueError("Transformation must either act on another Transformation or on a SimSnap")

        self.applied = False
        if not defer:
            self.apply(force=False)

    @property
    def sim(self):
        return self._sim()

    @sim.setter
    def sim(self, sim):
        if sim is None:
            self._sim = lambda: None
        else:
            self._sim = weakref.ref(sim)

    def apply_to(self, f):
        if self.next_transformation is not None:
            # this is a chained transformation, get the SimSnap to operate on
            # from the level below
            f = self.next_transformation.apply_to(f)

        self._apply(f)

        return f

    def apply_inverse_to(self, f):
        self._revert(f)

        if self.next_transformation is not None:
            self.next_transformation.apply_inverse_to(f)



    def apply(self, force=True):

        if self.next_transformation is not None:
            # this is a chained transformation, get the SimSnap to operate on
            # from the level below
            f = self.next_transformation.apply(force=force)
        else:
            f = self.sim

        if self.applied and force:
            raise RuntimeError("Transformation has already been applied")

        if not self.applied:
            self._apply(f)
            self.applied = True
            self.sim = f

        return f

    def revert(self):
        if not self.applied:
            raise RuntimeError("Transformation has not been applied")
        self._revert(self.sim)
        self.applied = False
        if self.next_transformation is not None:
            self.next_transformation.revert()

    def _apply(self, f):
        pass

    def _revert(self, f):
        pass

    def __enter__(self):
        self.apply(force=False)

    def __exit__(self, *args):
        self.revert()


class GenericTranslation(Transformation):

    def __init__(self, f, arname, shift):
        self.shift = shift
        self.arname = arname
        super().__init__(f)

    def _apply(self, f):
        f[self.arname] += self.shift

    def _revert(self, f):
        f[self.arname] -= self.shift


class GenericRotation(Transformation):

    def __init__(self, f, matrix, ortho_tol=1.e-8):
        # Check that the matrix is orthogonal
        resid = np.dot(matrix, np.asarray(matrix).T) - np.eye(3)
        resid = (resid ** 2).sum()
        if resid > ortho_tol or resid != resid:
            raise ValueError("Transformation matrix is not orthogonal")
        self.matrix = matrix
        super().__init__(f)

    def _apply(self, f):
        f._transform(self.matrix)

    def _revert(self, f):
        f._transform(self.matrix.T)


def translate(f, shift):
    """Form a context manager for translating the simulation *f* by the given
    spatial *shift*.

    This allows you to enclose a code block within which the simulation is offset
    by the specified amount. On exiting the code block, you are guaranteed the
    simulation is put back to where it started, so

    with translate(f, shift) :
        print f['pos'][0]

    is equivalent to

    try:
        f['pos']+=shift
        print f['pos'][0]
    finally:
        f['pos']-=shift

    On the other hand,

    translate(f, shift)
    print f['pos'][0]

    Performs the translation but does not revert it at any point.
    """
    return GenericTranslation(f, 'pos', shift)


def inverse_translate(f, shift):
    """Form a context manager for translating the simulation *f* by the spatial
    vector *-shift*.

    For a fuller description, see *translate*"""
    return translate(f, -np.asarray(shift))


def v_translate(f, shift):
    """Form a context manager for translating the simulation *f* by the given
    velocity *shift*.

    For a fuller description, see *translate* (which applies to position transformations)."""

    return GenericTranslation(f, 'vel', shift)


def inverse_v_translate(f, shift):
    """Form a context manager for translating the simulation *f* by the given
    velocity *-shift*.

    For a fuller description, see *translate* (which applies to position transformations)."""

    return GenericTranslation(f, 'vel', -np.asarray(shift))


def xv_translate(f, x_shift, v_shift):
    """Form a context manager for translating the simulation *f* by the given
    position *x_shift* and velocity *v_shift*.

    For a fuller description, see *translate* (which applies to position transformations)."""

    return translate(v_translate(f, v_shift),
                     x_shift)


def inverse_xv_translate(f, x_shift, v_shift):
    """Form a context manager for translating the simulation *f* by the given
    position *-x_shift* and velocity *-v_shift*.

    For a fuller description, see *translate* (which applies to position transformations)."""

    return translate(v_translate(f, -np.asarray(v_shift)),
                     -np.asarray(x_shift))


def transform(f, matrix):
    """Form a context manager for rotating the simulation *f* by the given 3x3
    *matrix*"""

    return GenericRotation(f, matrix)


def null(f):
    """Form a context manager for the null transformation (useful to avoid messy extra logic for situations where it's
    unclear whether any transformation will be applied)"""

    return Transformation(f)



"""
pynbody
=======

A light-weight, portable, format-transparent analysis framework
for N-body and SPH astrophysical simulations.

For more information, either build the latest documentation included
in our git repository, or view the online version here:
http://pynbody.github.io/pynbody/

"""

import sys

# We need to import configuration first, so prevent isort from reordering
# isort: off
from .configuration import config, logger, config_parser
# isort: on
from . import (
    analysis,
    array,
    bridge,
    configuration,
    derived,
    family,
    filt,
    gravity,
    halo,
    snapshot,
    sph,
    transformation,
    util,
)


# The PlotModuleProxy serves to delay import of pynbody.plot until it's accessed.
# Importing pynbody.plot imports pylab which in turn is quite slow and cause problem
# for terminal-only applications. However, since pynbody always auto-imported
# pynbody.plot, it would seem to be too destructive to stop this behaviour.
# So this hack is the compromise and should be completely transparent to most
# users.
class PlotModuleProxy:
    def _do_import(self):
        global plot
        del plot
        from . import plot as plot_module
        plot = plot_module

    def __hasattr__(self, key):
        self._do_import()
        return hasattr(plot, key)

    def __getattr__(self, key):
        self._do_import()
        global plot
        return getattr(plot, key)

    def __setattr__(self, key, value):
        self._do_import()
        global plot
        return setattr(plot, key, value)

    def __dir__(self):
        self._do_import()
        global plot
        return dir(plot)

    def __repr__(self):
        return "<Unloaded plot module>"

plot = PlotModuleProxy()

from .snapshot import load, new

derived_array = snapshot.simsnap.SimSnap.derived_quantity

__version__ = '2.0.0-beta.2'

__all__ = ['load', 'new', 'derived_array']



from __future__ import annotations

from typing import Iterable


class IterableSubclasses:
    """A mixin for classes where we need to be able to iterate over their subclasses, possibly in a
    user-specified order. This is used by HaloCatalogue and SimSnap to find a suitable loader for
    a given file. """
    @classmethod
    def iter_subclasses(cls) -> Iterable[type]:
        for c in cls.__subclasses__():
            yield from c.iter_subclasses()
            yield c

    @classmethod
    def iter_subclasses_with_priority(cls, priority: Iterable[str | type]) -> Iterable[type]:
        """Iterate over all subclasses, starting with the given priorities

        The priorities can be provided either as a string or a class"""
        all_subclasses = list(cls.iter_subclasses())
        all_subclasses_name = [s.__name__ for s in all_subclasses]
        for next_priority in priority:
            if isinstance(next_priority, type):
                if next_priority in all_subclasses:
                    index = all_subclasses.index(next_priority)
                    del all_subclasses[index]
                    del all_subclasses_name[index]
                yield next_priority
            else:
                if next_priority not in all_subclasses_name:
                    raise ValueError(f"Unknown class {next_priority}")
                next_priority_index = all_subclasses_name.index(next_priority)
                yield all_subclasses[next_priority_index]
                del all_subclasses[next_priority_index]
                del all_subclasses_name[next_priority_index]

        # now iterate the rest
        yield from all_subclasses



"""

cosmology
=========

A set of functions for common cosmological calculations.

"""

import math

import numpy as np
import scipy
import scipy.integrate
from scipy.interpolate import interp1d

from .. import units
from ..array import SimArray
from ..configuration import config_parser

_interp_points = int(config_parser.get('general','cosmo-interpolation-points'))


def _a_dot(a, h0, om_m, om_l):
    om_k = 1.0 - om_m - om_l
    return h0 * a * np.sqrt(om_m * (a ** -3) + om_k * (a ** -2) + om_l)


def _a_dot_recip(*args):
    return 1. / _a_dot(*args)


def _da_dtau(a, h0, om_m, om_l):
    return a**2 * _a_dot(a, h0, om_m, om_l)

def _da_dtau_recip(*args):
    return 1. / _da_dtau(*args)


def hzoverh0(a, omegam0):
    """ returns: H(a) / H0  = [omegam/a**3 + (1-omegam)]**0.5 """
    return np.sqrt(omegam0 * np.power(a, -3) + (1. - omegam0))


def _lingrowthintegrand(a, omegam0):
    """ (e.g. eq. 8 in lukic et al. 2008)   returns: da / [a*H(a)/H0]**3 """
    return np.power((a * hzoverh0(a, omegam0)), -3)


def _lingrowthfac(red, omegam0, omegal0, return_norm=False):
    """
    returns: linear growth factor, b(a) normalized to 1 at z=0, good for flat lambda only
    a = 1/1+z
    b(a) = Delta(a) / Delta(a=1)   [ so that b(z=0) = 1 ]
    (and b(a) [Einstein de Sitter, omegam=1] = a)

    Delta(a) = 5 omegam / 2 H(a) / H(0) * integral[0:a] [da / [a H(a) H0]**3]
    equation  from  peebles 1980 (or e.g. eq. 8 in lukic et al. 2008) """
# need to add w ~= , nonflat, -1 functionality

    import scipy.integrate

    if (abs(omegam0 + omegal0 - 1.) > 1.e-4):
        raise RuntimeError("Linear growth factors can only be calculated for flat cosmologies")

    a = 1 / (1. + red)

    # 1st calc. for z=z
    lingrowth = scipy.integrate.quad(_lingrowthintegrand, 0., a, (omegam0))[0]
    lingrowth *= 5. / 2. * omegam0 * hzoverh0(a, omegam0)

    # then calc. for z=0 (for normalization)
    a0 = 1.
    lingrowtha0 = scipy.integrate.quad(
        _lingrowthintegrand, 0., a0, (omegam0))[0]
    lingrowtha0 *= 5. / 2. * omegam0 * hzoverh0(a0, omegam0)

    lingrowthfactor = lingrowth / lingrowtha0
    if return_norm:
        return lingrowthfactor, lingrowtha0
    else:
        return lingrowthfactor


def linear_growth_factor(f, z=None):
    """Calculate the linear growth factor b(a), normalized to 1
    at z=0, for the cosmology of snapshot f.

    The output is dimensionless. If a redshift z is
    specified, it is used in place of the redshift in
    output f.
    """
    if z is None:
        z = f.properties['z']
    omegam0 = f.properties['omegaM0']
    omegal0 = f.properties['omegaL0']
    return _lingrowthfac(z, omegam0, omegal0)


def rate_linear_growth(f, z=None, unit='h Gyr^-1'):
    """Calculate the linear growth rate b'(a), normalized
    to 1 at z=0, for the cosmology of snapshot f.

    The output is in 'h Gyr^-1' by default. If a redshift z is specified,
    it is used in place of the redshift in output f."""

    if z is None:
        z = f.properties['z']
    a = 1. / (1. + z)
    omegam0 = f.properties['omegaM0']
    omegal0 = f.properties['omegaL0']

    b, X = _lingrowthfac(z, omegam0, omegal0, return_norm=True)
    I = _lingrowthintegrand(a, omegam0)

    term1 = -(1.5 * omegam0 * a ** -3) * b / \
        math.sqrt(1. - omegam0 + omegam0 * a ** -3)
    term2 = (2.5 * omegam0) * hzoverh0(a, omegam0) ** 2 * a * I / X

    res = units.h * (term1 + term2) * 100. * units.Unit("km s^-1 Mpc^-1")

    return res.in_units(unit, **f.conversion_context())


def _test_rate_linear_growth(f, z=None, unit='h Gyr^-1'):
    # coded up by AP to test linear growth *rate* equation above
    if z is None:
        z = f.properties['z']
    a0 = 1. / (1. + z)
    a1 = a0 * 0.999
    z0 = 1. / a0 - 1
    z1 = 1. / a1 - 1

    b0 = linear_growth_factor(f, z0)
    b1 = linear_growth_factor(f, z1)

    db = b1 - b0

    unit = units.Unit(unit)
    dt = age(f, z1, unit ** -1) - age(f, z0, unit ** -1)

    return db / dt


def age(f, z=None, unit='Gyr'):
    """
    Calculate the age of the universe in the snapshot f
    by integrating the Friedmann equation.

    The output is given in the specified units. If a redshift
    z is specified, it is used in place of the redshift in the
    output f.

    **Input**:

    *f*: SimSnap

    **Optional Keywords**:

    *z (None)*: desired redshift. Can be a single number, a list, or a
    np.ndarray.

    *unit ('Gyr')*: desired units for age output

    """
    if z is None:
        z = f.properties['z']

    h0 = f.properties['h']
    omM = f.properties['omegaM0']
    omL = f.properties['omegaL0']

    conv = units.Unit("0.01 s Mpc km^-1").ratio(unit, **f.conversion_context())

    def get_age(x):
        x = 1.0 / (1.0 + x)
        return scipy.integrate.quad(_a_dot_recip, 0, x, (h0, omM, omL))[0] * conv

    if isinstance(z, np.ndarray) or isinstance(z, list):
        if len(z) > _interp_points:
            a_vals = np.logspace(-3,0, _interp_points)
            z_vals = 1./a_vals-1.
            log_age_vals = np.log(age(f, z_vals))
            interp = interp1d(np.log(a_vals), log_age_vals, bounds_error=False)
            log_a_input = np.log(1./(1.+z))
            results = np.exp(interp(log_a_input))
        else:
            results = np.array([get_age(_z) for _z in z])
        results = results.view(SimArray)
        results.units = unit
        return results
    else:
        return get_age(z)

def tau(f, z=None):
    """
    Calculate the conformal time of the universe in the snapshot f
    by integrating the Friedmann equation.

    The output is given in the specified units. If a redshift
    z is specified, it is used in place of the redshift in the
    output f.

    **Input**:

    *f*: SimSnap

    **Optional Keywords**:

    *z (None)*: desired redshift. Can be a single number, a list, or a
    np.ndarray.
    """
    if z is None:
        z = f.properties['z']

    h0 = f.properties['h']
    omM = f.properties['omegaM0']
    omL = f.properties['omegaL0']

    @np.vectorize
    def get_tau(z):
        aexp = 1.0 / (1.0 + z)
        return scipy.integrate.quad(
            _da_dtau_recip,
            1,
            aexp,
            args=(h0, omM, omL)
        )[0]

    return get_tau(z)


@units.takes_arg_in_units((1, "Gyr"), context_arg=0)
def redshift(f, time):
    """
    Calculate the redshift given a snapshot and a time since Big Bang
    in Gyr.

    Uses scipy.optimize.newton to do the root finding if number of
    elements in the time array is less than 1000, otherwise uses a linear
    interpolation.


    **Input**:

    *f*: SimSnap with cosmological parameters defined

    *time*: time since the Big Bang in Gyr for which a redshift should
     be returned. float, list, or np.ndarray

    """

    from scipy.interpolate import interp1d
    from scipy.optimize import newton

    def func(x, sim, time):
        return age(sim, x) - time

    if isinstance(time, list) or isinstance(time, np.ndarray):
        if len(time) > _interp_points:
            zs = np.logspace(3, -10, _interp_points)
            ages = age(f, zs)
            i = interp1d(ages, zs)
            return i(time)
        else:
            return np.array([newton(func, 1, args=(f, x)) for x in time])
    else:
        return newton(func, 1, args=(f, time))


def rho_crit(f, z=None, unit=None):
    """Calculate the critical density of the universe in
    the snapshot f.

    z specifies the redshift. If z is none, the redshift of the
    provided snapshot is used.

    unit specifies the units of the returned density. If unit is None,
    the returned density will be in the units of
    f["mass"].units/f["pos"].units**3. If that unit cannot be calculated,
    the returned units are Msol kpc^-3 comoving.

    Note that you can get slightly confusing results if your
    simulation is in comoving units and you specify a different
    redshift z. Specifically, the physical density for the redshift
    you specify is calulated, but expressed as a comoving density *at
    the redshift of the snapshot*. This is intentional behaviour."""

    if z is None:
        z = f.properties['z']

    if unit is None:
        try:
            unit = f.dm["mass"].units / f.dm["pos"].units ** 3
        except units.UnitsException:
            unit = units.NoUnit()

    if hasattr(unit, "_no_unit"):
        unit = units.Unit("Msol kpc^-3 a^-3")

    omM = f.properties['omegaM0']
    omL = f.properties['omegaL0']
    h0 = f.properties['h']
    a = 1.0 / (1.0 + z)

    H_z = _a_dot(a, h0, omM, omL) / a
    H_z = units.Unit("100 km s^-1 Mpc^-1") * H_z

    rho_crit = (3 * H_z ** 2) / (8 * math.pi * units.G)

    return rho_crit.ratio(unit, **f.conversion_context())


def rho_M(f, z=None, unit=None):
    """Calculate the matter density of the universe in snapshot f.

    unit and z are used if not None, as by rho_crit. See also the note in
    rho_crit about confusion over comoving units in this case."""

    if z is None:
        z = f.properties['z']

    return f.properties['omegaM0'] * rho_crit(f, 0, unit) * (1.0 + z) ** 3


def H(f):
    """Calculate the Hubble parameter of the universe in snapshot f"""
    return f.properties['h'] * hzoverh0(f.properties['a'], f.properties['omegaM0']) * units.Unit("100 km s^-1 Mpc^-1")


def add_hubble(f):
    """Add the hubble flow to velocities in snapshot f"""

    f['vel'] += f['pos'] * H(f)


def comoving_to_physical(ar):
    """Given an array, modify it to be in physical units (remove any
    dependencies on a or aform)."""

    a_power = ar.units._power_of("a")
    aform_power = ar.units._power_of("aform")

    if a_power != 0:
        a = ar.sim.properties['a']
        ar *= a ** a_power
        ar /= units.Unit("a") ** a_power
    if aform_power != 0:
        aform = ar.sim['aform']
        ar *= aform ** aform_power
        ar /= units.Unit("aform")



"""

profile
=======

A set of classes and functions for making profiles of simulation
properties.

"""

import logging
import math
import pickle
import warnings
from time import process_time

import numpy as np

import pynbody

from .. import array, units, util

logger = logging.getLogger('pynbody.analysis.profile')


class Profile:

    """

    A basic profile class for arbitrary profiles. Stores information
    about bins etc.

    Made to work with the pynbody SimSnap instances. The constructor
    only generates the bins and figures out which particles belong in
    which bin.  Profiles are generated lazy-loaded when a given
    property is requested.

    **Input**:

    *sim* : a simulation snapshot - this can be any subclass of SimSnap

    **Optional Keywords**:

    *ndim* (default = 2): specifies whether it's a 2D or 3D profile - in the
                       2D case, the bins are generated in the xy plane

    *type* (default = 'lin'): specifies whether bins should be spaced linearly ('lin'),
                            logarithmically ('log') or contain equal numbers of
                            particles ('equaln')

    *rmin* (default = min(x)): minimum radial value to consider

    *rmax* (default = max(x)): maximum radial value to consider

    *nbins* (default = 100): number of bins

    *bins* : array like - predefined bin edges in units of binning quantity. If this
             keyword is set, the values of the keywords *type*, *nbins*, *rmin* and *rmax*
             will be ignored

    *calc_x* (default = None): function to use to calculate the value
     for binning. If None it defaults to the radial distance from
     origin (in either 2 or 3 dimensions), ut you can specify this
     function to return any value you want for making profiles along
     arbitrary axes. Depening on your function, the units of certain
     profiles (such as density) might not make sense.

    *weight_by* (default = 'mass'): name of the array to use for weighting
     averages across particles in each bin

    **Output**:

    a Profile object. To find out which profiles are available, use keys().

    **Implemented profile functions**:

    *density*    : density

    *mass*       : mass in each bin

    *mass_enc*   : enclosed mass

    *fourier* : provides fourier coefficients, amplitude and phase for
     m=0 to m=6.  To access the amplitude profile of m=2 mode, do
     ``p['fourier']['amp'][2,:]``

    *dyntime*    : dynamical time

    *g_spherical*: GM_enc/r^2

    *rotation_curve_spherical*: rotation curve from vc = sqrt(GM/R) -
     can be very wrong!

    *j_circ*     : angular momentum of particles on circular orbits

    *v_circ* : circular velocity, aka rotation curve - calculated from
     the midplane gravity, so this can be expensive

    *E_circ*     : energy of particles on circular orbits in the midplane

    *omega*      : circular orbital frequency

    *kappa*      : radial orbital frequency

    *beta*       : 3-D velocity anisotropy parameter

    *magnitudes* : magnitudes in each bin - default band = 'v'

    *sb*         : surface brightness - default band = 'v'


    Additional functions should use the profile_property to yield the
    desired profile.

    **Lazy-loading arrays:**

    The Profile class will automatically compute a mass-weighted
    profile for any lazy-loadable array of its parent SimSnap object.

    **Dispersions:**

    To obtain a dispersion profile, attach a ``_disp`` after the desired
    quantity name.

    **RMS:**

    The root-mean-square of a quantity can be obtained by using a ``_rms`` suffix

    **Derivatives:**

    To compute a derivative of a profile, prepend a ``d_`` to the
    profile string, as in ``p['d_temp']`` to get a temperature gradient.

    **Saving and loading previously generated profiles:**

    Use the :func:`~pynbody.analysis.profile.Profile.write` function
    to write the current profiles with all the necessary information
    to a file. Initialize a profile with the load_from_file=True
    keyword to automatically load a previously saved profile. The
    filename is chosen automatically and corresponds to a hash
    generated from the positions of the particles used in the
    profile. This is to ensure that you are always looking at the same
    set of particles, centered in the same way. It also means you
    *must* use the same centering method if you want to reuse a saved
    profile.


    """

    _profile_registry = {}

    def _calculate_x(self, sim):
        return ((sim['pos'][:, 0:self.ndim] ** 2).sum(axis=1)) ** (1, 2)

    def __init__(self, sim, load_from_file=False, ndim=2, type='lin', calc_x=None, weight_by='mass', **kwargs):

        generate_new = True
        if calc_x is None:
            calc_x = self._calculate_x
        self.sim = sim
        self.type = type
        self.ndim = ndim
        self._weight_by = weight_by
        self._x = calc_x(sim)
        x = self._x

        if load_from_file:

            filename = self._get_unique_filepath_from_particle_list()

            try:
                with open(filename, "rb") as f:
                    data = pickle.load(f)
                self._properties = data['properties']
                self.max = data['max']
                self.min = data['min']
                self.nbins = data['nbins']
                self._profiles = data['profiles']
                self.binind = data['binind']

                logger.info("Loaded profile from %s" % filename)

                generate_new = False

            except FileNotFoundError:
                logger.warning(
                    "Existing profile not found -- generating one from scratch instead")

        if generate_new:
            self._properties = {}
            # The profile object is initialized given some array of values
            # and optional keyword parameters

            if 'max' in kwargs:
                kwargs['rmax'] = kwargs.pop('max')
                warnings.warn("Use of max as a keyword argument is deprecated. Use rmax instead.", DeprecationWarning)
            if 'min' in kwargs:
                kwargs['rmin'] = kwargs.pop('min')
                warnings.warn("Use of min as a keyword argument is deprecated. Use rmin instead.", DeprecationWarning)

            if 'rmax' in kwargs:
                if isinstance(kwargs['rmax'], str):
                    self.max = units.Unit(kwargs['rmax']).ratio(x.units,
                                                               **sim.conversion_context())
                else:
                    self.max = kwargs['rmax']
            else:
                self.max = np.max(x)
            if 'bins' in kwargs:
                self.nbins = len(kwargs['bins']) - 1
            elif 'nbins' in kwargs:
                self.nbins = kwargs['nbins']
            else:
                self.nbins = 100

            if 'rmin' in kwargs:
                if isinstance(kwargs['rmin'], str):
                    self.min = units.Unit(kwargs['rmin']).ratio(x.units,
                                                               **sim.conversion_context())
                else:
                    self.min = kwargs['rmin']
            else:
                self.min = np.min(x[x > 0])

            if 'bins' in kwargs:
                self._properties['bin_edges'] = kwargs['bins']
                self.min = kwargs['bins'].min()
                self.max = kwargs['bins'].max()
            elif type == 'log':
                self._properties['bin_edges'] = np.logspace(
                    np.log10(self.min), np.log10(self.max), num=self.nbins + 1)
            elif type == 'lin':
                self._properties['bin_edges'] = np.linspace(
                    self.min, self.max, num=self.nbins + 1)
            elif type == 'equaln':
                self._properties['bin_edges'] = util.equipartition(
                    x, self.nbins, self.min, self.max)
            else:
                raise RuntimeError("Bin type must be one of: lin, log, equaln")

            self['bin_edges'] = array.SimArray(self['bin_edges'], x.units)
            self['bin_edges'].sim = self.sim

            n, bins = np.histogram(self._x, self['bin_edges'])
            self._setup_bins()

            # set up the empty list of profiles
            self._profiles = {'n': n}

    def _setup_bins(self):
        # middle of the bins for convenience

        self._properties['rbins'] = 0.5 * (self['bin_edges'][:-1] +
                                           self['bin_edges'][1:])

        # no idea why this next line was there... for conversion_context
        # self['rbins'].sim = self.sim

        # Width of the bins
        self._properties['dr'] = np.gradient(
            self['rbins']).view(array.SimArray)
        # be extra cautious carrying over stuff because sometimes fails
        self._properties['dr'].units = self['rbins'].units
        self._properties['dr'].sim = self.sim

        self.binind = []
        if len(self._x) > 0:
            self.partbin = np.digitize(self._x, self['bin_edges'])
        else:
            self.partbin = np.array([])

        assert self.ndim in [2, 3]
        if self.ndim == 2:
            self._binsize = np.pi * (self['bin_edges'][1:] ** 2 -
                                     self['bin_edges'][:-1] ** 2)
        else:
            self._binsize = 4. / 3. * np.pi * (self['bin_edges'][1:] ** 3 -
                                               self['bin_edges'][:-1] ** 3)

        # sort the partbin array
        from bisect import bisect
        sortind = self.partbin.argsort()
        sort_pind = self.partbin[sortind]

        # create the bin index arrays
        prev_index = bisect(sort_pind, 0)
        for i in range(self.nbins):
            new_index = bisect(sort_pind, i + 1)
            self.binind.append(np.sort(sortind[prev_index:new_index]))
            prev_index = new_index

    def __len__(self):
        """Returns the number of bins used in this profile object"""
        return self.nbins

    def _get_profile(self, name):
        """Return the profile of a given kind"""
        x = name.split(",")
        if name in self._profiles:
            return self._profiles[name]

        elif x[0] in Profile._profile_registry:
            args = x[1:]
            self._profiles[name] = Profile._profile_registry[x[0]](self, *args)

            try:
                self._profiles[name].sim = self.sim
            except AttributeError:
                pass
            return self._profiles[name]

        elif name in list(self.sim.keys()) or name in self.sim.all_keys():
            self._profiles[name] = self._auto_profile(name)
            self._profiles[name].sim = self.sim
            return self._profiles[name]

        elif name[-5:] == "_disp" and (name[:-5] in list(self.sim.keys()) or name[:-5] in self.sim.all_keys()):
            logger.info("Auto-deriving %s" % name)
            self._profiles[name] = self._auto_profile(
                name[:-5], dispersion=True)
            self._profiles[name].sim = self.sim
            return self._profiles[name]

        elif name[-4:] == "_rms" and (name[:-4] in list(self.sim.keys()) or name[:-4] in self.sim.all_keys()):
            logger.info("Auto-deriving %s" % name)
            self._profiles[name] = self._auto_profile(name[:-4], rms=True)
            self._profiles[name].sim = self.sim
            return self._profiles[name]

        elif name[-4:] == "_med" and (name[:-4] in list(self.sim.keys()) or name[:-4] in self.sim.all_keys()):
            logger.info("Auto-deriving %s" % name)
            self._profiles[name] = self._auto_profile(name[:-4], median=True)
            self._profiles[name].sim = self.sim
            return self._profiles[name]

        elif name[0:2] == "d_" and (name[2:] in list(self.keys()) or name[2:] in self.derivable_keys() or name[2:] in self.sim.all_keys()):
            #            if np.diff(self['dr']).all() < 1e-13 :
            logger.info("Auto-deriving %s/dR" % name)
            self._profiles[name] = np.gradient(self[name[2:]], self['dr'][0])
            self._profiles[name] = self._profiles[name] / self['dr'].units
            return self._profiles[name]
            # else :
            #    raise RuntimeError, "Derivatives only possible for profiles of fixed bin width."

        else:
            raise KeyError(name + " is not a valid profile")

    def _auto_profile(self, name, dispersion=False, rms=False, median=False):
        result = np.zeros(self.nbins)

        # force derivation of array if necessary:
        self.sim[name]

        for i in range(self.nbins):
            subs = self.sim[self.binind[i]]
            name_array = subs[name].view(np.ndarray)
            mass_array = subs[self._weight_by].view(np.ndarray)

            if dispersion:
                sq_mean = (name_array ** 2 * mass_array).sum() / \
                    self['weight_fn'][i]
                mean_sq = (
                    (name_array * mass_array).sum() / self['weight_fn'][i]) ** 2
                try:
                    result[i] = math.sqrt(sq_mean - mean_sq)
                except ValueError:
                    # sq_mean<mean_sq occasionally from numerical roundoff
                    result[i] = 0

            elif rms:
                result[i] = np.sqrt(
                    (name_array ** 2 * mass_array).sum() / self['weight_fn'][i])
            elif median:
                if len(subs) == 0:
                    result[i] = np.nan
                else:
                    sorted_name = sorted(name_array)
                    result[i] = sorted_name[int(np.floor(0.5 * len(subs)))]
            else:
                result[i] = (name_array * mass_array).sum() / self['weight_fn'][i]

        result = result.view(array.SimArray)
        result.units = self.sim[name].units
        result.sim = self.sim
        return result

    def __getitem__(self, name):
        """Return the profile of a given kind"""
        if name in self._properties:
            return self._properties[name]
        else:
            return self._get_profile(name)

    def __setitem__(self, name, item):
        """Set the profile or property by hand"""
        if name in self._properties:
            self._properties[name] = item
        elif name in self._profiles:
            self._profiles[name] = item
        else:
            raise KeyError(name + " is not a valid profile or property")

    def __delitem__(self, name):
        del self._profiles[name]

    def __repr__(self):
        return ("<Profile: " +
                str(self.families()) + " ; " +
                str(self.ndim) + "D ; " +
                self.type) + " ; " + str(list(self.keys())) + ">"

    def keys(self):
        """Returns a listing of available profile types"""
        return list(self._profiles.keys())

    def derivable_keys(self):
        """Returns a list of possible profiles"""
        return list(self._profile_registry.keys())

    def families(self):
        """Returns the family of particles used"""
        return self.sim.families()

    def create_particle_array(self, profile_name, particle_name=None, out_sim=None):
        """Create a particle array with the results of the profile
        calculation.

        After calling this function, sim[particle_name][i] ==
        profile[profile_name][bin_in_which_particle_i_sits]

        If particle_name is not specified, it defaults to the same as
        profile_name.
        """
        import scipy
        import scipy.interpolate

        if particle_name is None:
            particle_name = profile_name

        if out_sim is None:
            out_sim = self.sim
            out_x = self._x
        else:
            out_x = self._calculate_x(out_sim)

        # nearest-neighbour version if interpolation unavailable
        #px = np.digitize(out_x, self.bins)-1
        #ok = np.where((px>=0) * (px<len(self)))
        #out_sim[particle_name, ok[0]] = self[profile_name][px[ok]]

        in_y = self[profile_name]

        if in_y.min() > 0:
            use_log = True
            in_y = np.log(in_y)
        else:
            use_log = False

        interp = scipy.interpolate.interp1d(np.log10(self.r), in_y, 'linear', copy=False,
                                            bounds_error=False)
        rep = interp(np.log10(out_x))

        if use_log:
            out_sim[particle_name] = np.exp(rep)
        else:
            out_sim[particle_name] = rep

        rep[np.where(out_x > math.log10(self.r.max()))] = self[
            profile_name][-1]
        rep[np.where(out_x < math.log10(self.r.min()))] = self[profile_name][0]

        out_sim[particle_name].units = self[profile_name].units

    def _generate_hash_filename_from_particles(self):
        """Create a filename for the saved profile from a hash using the binning data"""

        import hashlib

        # Changing to the new() method, which will not fail if usedforsecurity is unsupported
        # in a given system configuration (issue 581)
        h = hashlib.new('md5')
        # Reproduce old behaviour, given byte-like data to create the hash.
        h.update(self._x)
        return h.hexdigest()

    def _get_unique_filepath_from_particle_list(self):

        try:
            folder_path = self.sim.base.filename
        except AttributeError:
            folder_path = self.sim.filename

        unique_hash = self._generate_hash_filename_from_particles()
        filename = folder_path + '.profile.' + unique_hash
        return filename

    def write(self):
        """
        Writes all the vital information of the profile to a file.

        To recover the profile, initialize a profile with the
        load_from_file=True keyword to automatically load a previously
        saved profile. The filename is chosen automatically and
        corresponds to a hash generated from the positions of the
        particles used in the profile. This is to ensure that you are
        always looking at the same set of particles, centered in the
        same way. It also means you *must* use the same centering
        method if you want to reuse a saved profile.


        """

        # record all the important data except for the snapshot itself
        # use the hash generated from the particle list for the file name
        # suffix

        filename = self._get_unique_filepath_from_particle_list()

        logger.info("Writing profile to %s", filename)

        with open(filename, "wb") as f:
            pickle.dump(
                {
                    'properties': self._properties,
                    'max': self.max,
                    'min': self.min,
                    'nbins': self.nbins,
                    'profiles': self._profiles,
                    'binind': self.binind
                },
                f,
            )

    @staticmethod
    def profile_property(fn):
        Profile._profile_registry[fn.__name__] = fn
        return fn


@Profile.profile_property
def weight_fn(self, weight_by=None):
    """
    Calculate mass in each bin
    """
    if weight_by is None:
        weight_by = self._weight_by
    mass = array.SimArray(np.zeros(self.nbins), self.sim[weight_by].units)

    with self.sim.immediate_mode:
        pmass = self.sim[weight_by].view(np.ndarray)

    for i in range(self.nbins):
        mass[i] = (pmass[self.binind[i]]).sum()

    mass.sim = self.sim
    mass.units = self.sim[weight_by].units

    return mass

@Profile.profile_property
def mass(self):
    return weight_fn(self, 'mass')



@Profile.profile_property
def density(self):
    """
    Generate a radial density profile for the current type of profile
    """
    return self['mass'] / self._binsize


@Profile.profile_property
def fourier(self, delta_t = "0.1 Myr", phi_bins=100):
    """
    Generate a profile of fourier coefficients, amplitudes and phases
    """

    delta_t = pynbody.units.Unit(delta_t)


    f = {'c': np.zeros((7, self.nbins), dtype=complex),
         'c_delta': np.zeros((7, self.nbins), dtype=complex),
         'amp': np.zeros((7, self.nbins)),
         'phi': np.zeros((7, self.nbins)),
         'dphi_dt': np.zeros((7, self.nbins))}

    for i in range(self.nbins):
        if self._profiles['n'][i] > 100:
            phi = np.arctan2(
                self.sim['y'][self.binind[i]], self.sim['x'][self.binind[i]])
            mass = self.sim['mass'][self.binind[i]]

            x1 = self.sim['x'][self.binind[i]] + self.sim['vx'][self.binind[i]] * delta_t
            y1 = self.sim['y'][self.binind[i]] + self.sim['vy'][self.binind[i]] * delta_t
            phi1 = np.arctan2(y1,x1)

            hist_range = (-np.pi, np.pi)
            hist, binphi = np.histogram(phi, weights=mass, bins=phi_bins, range=(hist_range))
            hist1, _ = np.histogram(phi1, weights=mass, bins=phi_bins, range=(hist_range))
            binphi = binphi[:-1] + .5*np.diff(binphi)
            for m in range(7):
                f['c'][m, i] = np.sum(hist * np.exp(-1j * m * binphi))
                f['c_delta'][m, i] = np.sum(hist1 * np.exp(-1j * m * binphi))

    f['c'][:, self['mass'] > 0] /= self['mass'][self['mass'] > 0]
    f['amp'] = np.sqrt(np.imag(f['c']) ** 2 + np.real(f['c']) ** 2)
    f['phi'] = np.arctan2(np.imag(f['c']), np.real(f['c']))

    dphi = np.arctan2(np.imag(f['c_delta']), np.real(f['c_delta'])) - f['phi']
    dphi[dphi>np.pi] = dphi[dphi>np.pi]-2*np.pi
    dphi[dphi<-np.pi] = dphi[dphi<-np.pi]+2*np.pi
    dphi = array.SimArray(dphi,"1")
    f['dphi_dt'] = (dphi / delta_t).in_units(self.sim['vx'].units/self.sim['x'].units)

    return f

@Profile.profile_property
def pattern_frequency(pro):
    """Estimate the pattern speed from the m=2 Fourier mode"""
    return pro['fourier']['dphi_dt'][2,:]/2

@Profile.profile_property
def mass_enc(self):
    """
    Generate the enclosed mass profile
    """
    return self['mass'].cumsum()


@Profile.profile_property
def density_enc(self):
    """
    Generate the mean enclosed density profile
    """
    return self['mass_enc'] / ((4. * math.pi / 3) * self['rbins'] ** 3)


@Profile.profile_property
def dyntime(self):
    """The dynamical time of the bin, sqrt(R^3/2GM)."""
    dyntime = (self['rbins'] ** 3 / (2 * units.G * self['mass_enc'])) ** (1, 2)
    return dyntime


@Profile.profile_property
def g_spherical(self):
    """The naive gravitational acceleration assuming spherical
    symmetry = GM_enc/r^2"""

    return (units.G * self['mass_enc'] / self['rbins'] ** 2)


@Profile.profile_property
def rotation_curve_spherical(self):
    """
    The naive rotation curve assuming spherical symmetry: vc = sqrt(G M_enc/r)
    """

    # .in_units('km s**-1')
    return ((units.G * self['mass_enc'] / self['rbins']) ** (1, 2))


@Profile.profile_property
def j_circ(p):
    """Angular momentum of particles on circular orbits."""
    return p['v_circ'] * p['rbins']


@Profile.profile_property
def v_circ(p, grav_sim=None):
    """Circular velocity, i.e. rotation curve. Calculated by computing the gravity
    in the midplane - can be expensive"""

    import pynbody.gravity.calc as gravity

    from .. import config

    global config

    grav_sim = grav_sim or p.sim

    logger.warning(
        "Profile v_circ -- this routine assumes the disk is in the x-y plane")

    # If this is a cosmological run, go up to the halo level
    # if hasattr(grav_sim,'base') and grav_sim.base.properties.has_key("halo_number") :
    #    while hasattr(grav_sim,'base') and grav_sim.base.properties.has_key("halo_number") :
    #        grav_sim = grav_sim.base

    # elif hasattr(grav_sim,'base') :
    #    grav_sim = grav_sim.base

    start = process_time()
    rc = gravity.midplane_rot_curve(
        grav_sim, p['rbins']).in_units(p.sim['vel'].units)
    end = process_time()
    logger.info("Rotation curve calculated in %5.3g s" % (end - start))
    return rc


@Profile.profile_property
def E_circ(p):
    """Energy of particles on circular orbits."""
    return 0.5 * (p['v_circ'] ** 2) + p['pot']


@Profile.profile_property
def pot(p):
    """Calculates the potential in the midplane - can be expensive"""
    #from . import gravity
    import pynbody.gravity.calc as gravity

    logger.warning(
        "Profile pot -- this routine assumes the disk is in the x-y plane")

    grav_sim = p.sim
    # Go up to the halo level
    while hasattr(grav_sim, 'base') and "halo_number" in grav_sim.base.properties:
        grav_sim = grav_sim.base

    start = process_time()
    pot = gravity.midplane_potential(
        grav_sim, p['rbins']).in_units(p.sim['vel'].units ** 2)
    end = process_time()
    logger.info("Potential calculated in %5.3g s" % (end - start))
    return pot


@Profile.profile_property
def omega(p):
    """Circular frequency Omega = v_circ/radius (see Binney & Tremaine Sect. 3.2)"""
    prof = p['v_circ'] / p['rbins']
    prof.set_units_like('km s**-1 kpc**-1')
    return prof


@Profile.profile_property
def kappa(p):
    """Radial frequency kappa = sqrt(R dOmega^2/dR + 4 Omega^2) (see Binney & Tremaine Sect. 3.2)"""
    dOmega2dR = np.gradient(p['omega'] ** 2)/np.gradient(p['rbins'])
    return np.sqrt(p['rbins'] * dOmega2dR + 4 * p['omega'] ** 2)


@Profile.profile_property
def beta(p):
    """3D Anisotropy parameter as defined in Binney and Tremiane"""
    assert p.ndim == 3
    return 1.5 - (p['vx_disp'] ** 2 + p['vy_disp'] ** 2 + p['vz_disp'] ** 2) / p['vr_disp'] ** 2 / 2.


@Profile.profile_property
def magnitudes(self, band='v'):
    """
    Calculate magnitudes in each bin
    """
    from . import luminosity

    magnitudes = np.zeros(self.nbins)
    for i in range(self.nbins):
        magnitudes[i] = luminosity.halo_mag(
            self.sim[self.binind[i]], band=band)
    magnitudes = array.SimArray(magnitudes, units.Unit('1'))
    magnitudes.sim = self.sim
    return magnitudes


@Profile.profile_property
def sb(self, band='v'):
    # At 10 pc (distance for absolute magnitudes), 1 arcsec is 10 AU=1/2.06e4 pc
    # In [5]: (np.tan(np.pi/180/3600)*10.0)**2
    # Out[5]: 2.3504430539466191e-09
    # 1 square arcsecond is thus 2.35e-9 pc^2
    sqarcsec_in_bin = self._binsize.in_units('pc^2') / 2.3504430539466191e-09
    bin_luminosity = 10.0 ** (-0.4 * self['magnitudes,' + band])
    #import pdb; pdb.set_trace()
    surfb = -2.5 * np.log10(bin_luminosity / sqarcsec_in_bin)
    surfb = array.SimArray(surfb, units.Unit('1'))
    surfb.sim = self.sim
    return surfb


@Profile.profile_property
def Q(self):
    """Toomre Q parameter"""
    return (self['vr_disp'] * self['kappa'] / (3.36 * self['density'] * units.G)).in_units("")


@Profile.profile_property
def X(self):
    """X parameter defined as kappa^2*R/(2*pi*G*sigma*m)
    See Binney & Tremaine 2008, eq. 6.77"""

    lambda_crit = 4. * np.pi ** 2 * units.G * \
        self['density'] / (self['kappa'] ** 2)
    kcrit = 2. * np.pi / lambda_crit
    return (kcrit * self['rbins'] / 2).in_units("")


@Profile.profile_property
def jtot(self):
    """
    Magnitude of the total angular momentum
    """
    jtot = np.zeros(self.nbins)

    for i in range(self.nbins):
        subs = self.sim[self.binind[i]]
        jx = (subs['j'][:, 0] * subs['mass']).sum() / self['mass'][i]
        jy = (subs['j'][:, 1] * subs['mass']).sum() / self['mass'][i]
        jz = (subs['j'][:, 2] * subs['mass']).sum() / self['mass'][i]

        jtot[i] = np.sqrt(jx ** 2 + jy ** 2 + jz ** 2)

    return jtot


@Profile.profile_property
def j_theta(self):
    """
    Angle that the angular momentum vector of the bin makes with respect to the xy-plane.
    """

    return np.arccos(self['jz'] / self['jtot'])


@Profile.profile_property
def j_phi(self):
    """
    Angle that the angular momentum vector of the bin makes with the x-axis in the xy plane.
    """
    j_phi = np.zeros(self.nbins)

    for i in range(self.nbins):
        subs = self.sim[self.binind[i]]
        jx = (subs['j'][:, 0] * subs['mass']).sum() / self['mass'][i]
        jy = (subs['j'][:, 1] * subs['mass']).sum() / self['mass'][i]
        j_phi[i] = np.arctan(jy, jx)

    return j_phi


class InclinedProfile(Profile):

    """

    Creates a profile object to be used with a snapshot inclined by
    some known angle to the xy plane.

    In addition to the SimSnap object, it also requires the angle to
    initialize.

    **Example:**

    >>> s = pynbody.load('sim')
    >>> pynbody.analysis.angmom.faceon(s)
    >>> s.rotate_x(60)
    >>> p = pynbody.profile.InclinedProfile(s, 60)

    """

    def _calculate_x(self, sim):
        # calculate an ellipsoidal radius
        return (sim['x'] ** 2 + (sim['y'] / np.cos(np.radians(self.angle))) ** 2) ** (1, 2)

    def __init__(self, sim, angle, load_from_file=False, ndim=2, type='lin', **kwargs):
        self.angle = angle

        Profile.__init__(
            self, sim, load_from_file=load_from_file, ndim=ndim, type=type, **kwargs)

        # define the minor axis
        self._properties['rbins_min'] = np.cos(
            np.radians(self.angle)) * self['rbins']


class VerticalProfile(Profile):

    """

    Creates a profile object that uses the absolute value of the z-coordinate for binning.

    **Input**:

    *sim*: snapshot to make a profile from

    *rmin*: minimum radius for particle selection in kpc

    *rmax*: maximum radius for particle selection in kpc

    *zmax*: maximum height to consider in kpc

    **Optional Keywords**:

    *ndim*: if ndim=2, an edge-on projected profile is produced,
     i.e. density is in units of mass/pc^2. If ndim=3 a volume
     profile is made, i.e. density is in units of mass/pc^3.

    """

    def _calculate_x(self, sim):
        return array.SimArray(np.abs(sim['z']), sim['z'].units)

    def __init__(self, sim, rmin, rmax, zmax, load_from_file=False, ndim=3, type='lin', **kwargs):

        if isinstance(rmin, str):
            rmin = units.Unit(rmin)
        if isinstance(rmax, str):
            rmax = units.Unit(rmax)
        if isinstance(zmax, str):
            zmax = units.Unit(zmax)
        self.rmin = rmin
        self.rmax = rmax
        self.zmax = zmax

        # create a snapshot that only includes the section of disk we're
        # interested in
        assert ndim in [2, 3]
        if ndim == 3:
            sub_sim = sim[
                pynbody.filt.Disc(rmax, zmax) & ~pynbody.filt.Disc(rmin, zmax)]
        else:
            sub_sim = sim[(pynbody.filt.BandPass('x', rmin, rmax) |
                           pynbody.filt.BandPass('x', -rmax, -rmin)) &
                          pynbody.filt.BandPass('z', -zmax, zmax)]

        Profile.__init__(
            self, sub_sim, load_from_file=load_from_file, ndim=ndim, type=type, **kwargs)

    def _setup_bins(self):
        Profile._setup_bins(self)

        dr = self.rmax - self.rmin

        if self.ndim == 2:
            self._binsize = (
                self['bin_edges'][1:] - self['bin_edges'][:-1]) * dr
        else:
            area = array.SimArray(
                np.pi * (self.rmax ** 2 - self.rmin ** 2), "kpc^2")
            self._binsize = (
                self['bin_edges'][1:] - self['bin_edges'][:-1]) * area


class QuantileProfile(Profile):

    """

    Creates a profile object that returns the requested quantiles
    for a given array in a given bin.  The quantiles may be mass weighted.

    **Input**:

    *sim*: snapshot to make a profile from

    *q (default: (0.16,0.5,0.84))*:
             The quantiles that will be returned.
             Default is median with 1-sigma on either side.
             q can be of arbitrary length allowing the user to select
             any quantiles they desire.

    *weights (default:None)*:
             What should be used to weight the quantile.  You will usually
             want to use particle mass: sim['mass'].
             The default is to not weight by anything, weights=None.

    **Optional Keywords**:

    *ndim*: if ndim=2, an edge-on projected profile is produced,
     i.e. density is in units of mass/pc^2. If ndim=3 a volume
     profile is made, i.e. density is in units of mass/pc^3.

    """

    def __init__(self, sim, q=(0.16, 0.50, 0.84), weights=None, load_from_file = False, ndim = 3, type = 'lin', **kwargs):

        # create a snapshot that only includes the section of disk we're
        # interested in
        self.quantiles = q
        self.qweights = weights

        Profile.__init__(
            self, sim, load_from_file=load_from_file, ndim=ndim, type=type, **kwargs)

    def _get_profile(self, name):
        """Return the profile of a given kind"""
        x = name.split(",")
        if name in self._profiles:
            return self._profiles[name]

        elif name in list(self.sim.keys()) or name in self.sim.all_keys():
            self._profiles[name] = self._auto_profile(name)
            self._profiles[name].sim = self.sim
            return self._profiles[name]

        else:
            raise KeyError(name + " is not a valid QuantileProfile")

    def _auto_profile(self, name, dispersion=False, rms=False, median=False):
        result = np.zeros((self.nbins, len(self.quantiles)))
        for i in range(self.nbins):
            subs = self.sim[self.binind[i]]
            with self.sim.immediate_mode:
                name_array = subs[name].view(np.ndarray)
                sorted_array = sorted(name_array)
                topind = len(name_array) - 1
                if self.qweights is not None:
                    sorted_weights = self.qweights[np.argsort(name_array)]

            for iq, q in enumerate(self.quantiles):
                #import pdb; pdb.set_trace()
                if len(name_array) > 0:
                    if self.qweights is None:
                        ilow = int(np.floor(q * topind))
                        inc = q * topind - ilow
                        lowval = sorted_array[ilow]
                        hival = sorted_array[ilow + 1]
                        result[i, iq] = lowval + inc * (hival - lowval)
                    else:
                        cumw = np.cumsum(
                            sorted_weights) / np.sum(sorted_weights)
                        imin = min(
                            np.arange(len(sorted_array)), key=lambda x: abs(cumw[x] - q))
                        inc = q - cumw[imin]
                        lowval = sorted_array[imin]
                        if inc > 0:
                            nextval = sorted_array[imin + 1]
                        else:
                            if imin == 0:
                                nextval = lowval
                            else:
                                nextval = sorted_array[imin - 1]

                        result[i, iq] = lowval + inc * (nextval - lowval)
                        #if (result[i,iq] < 1e-6) : import pdb; pdb.set_trace()
                else:
                    result[i, iq] = np.nan
                    self['rbins'][i] = np.nan

        result = result.view(array.SimArray)
        result.units = self.sim[name].units
        result.sim = self.sim
        return result



import numpy.testing as npt

import pynbody
from pynbody import units


def numacc(a, b, tol=1.e-9):
    print(a, b)
    assert abs(a - b) < a * tol


def test_units_conversion():

    numacc(units.kpc.in_units(units.Mpc), 0.001)
    numacc(units.Mpc.in_units(units.kpc), 1000)
    numacc(units.yr.in_units(units.Myr), 1.e-6)
    numacc(units.au.in_units(units.pc), 4.84813681e-6)


def test_units_manipulation():
    # Just do some manipulation and check it's happy
    (units.kpc * units.yr) ** (1, 3) / units.Myr
    (units.a * units.erg) ** 9


def test_units_substitution():
    numacc((units.a / units.h).in_units(units.Unit(""), a=22, h=2), 11)


def test_units_parser():
    testunit = units.Unit("kpc a s^-2/3 Myr^2/3")
    print("Unit as parsed: ", testunit)
    testunit /= units.kpc
    testunit /= units.a
    testunit /= units.s ** (-2, 3)
    testunit /= units.Myr ** (2, 3)
    print("This should be one: ", testunit)
    assert abs(testunit.dimensionless_constant() - 1) < 1.e-10


def test_units_copy():
    # These should succeed

    import copy
    copy.copy(units.Unit("Msol kpc^-1")).ratio("kg km^-1")
    copy.deepcopy(units.Unit("Msol kpc^-1")).ratio("kg km^-1")


def test_units_pickle():
    import pickle
    pick = lambda x: pickle.loads(pickle.dumps(x))

    assert pick(units.km) is units.km  # named
    assert pick(units.m) is units.m  # irreducible
    assert pick(units.Unit("km s^-1 Msol^-5")) == units.Unit("km s^-1 Msol^-5")

def test_units_rdiv():
    assert 4.0/pynbody.units.m_p == pynbody.units.Unit("4.0 m_p^-1")

def test_dimensionless_addition():
    dimless_unit = units.Unit("0.5")
    print(dimless_unit-0.25)
    npt.assert_allclose(float(dimless_unit-0.25),0.25)
    npt.assert_allclose(float(dimless_unit+0.25),0.75)

def test_nounit_equality():
    assert units.Unit("1") == units.Unit("1")
    assert units.Unit("1") != units.NoUnit()
    assert units.NoUnit() == units.NoUnit()
def test_unit_from_numeric():
    assert units.Unit(1) == units.Unit("1")
    assert units.Unit(0.5) == units.Unit("0.5")
def test_units_addition():
    _2_kpc = units.Unit("2.0 kpc")
    _3_Mpc = units.Unit("3.0 Mpc")
    npt.assert_allclose((_2_kpc + _3_Mpc).in_units("kpc"),3002)
    npt.assert_allclose((_3_Mpc - _2_kpc).in_units("kpc"), 2998)
    npt.assert_allclose((_2_kpc + 2.0).in_units("kpc"), 4.0)
    npt.assert_allclose((_3_Mpc + 2.0).in_units("kpc"), 5000)

def test_units_zero_equality():
    assert units.Unit("0.0 km s^-1") == units.Unit("0.0 cm s^-1")

def test_units_equality():
    assert units.Unit("1.0 km s^-1") == units.Unit("1e5 cm s^-1")

def test_get_with_unit():
    ar1d = pynbody.array.SimArray([1.,2.,3.], units="km")
    ar2d = pynbody.array.SimArray([[1., 2., 3.], [4., 5., 6.]], units="km")
    assert pynbody.units.get_item_with_unit(ar1d, 1) == pynbody.units.Unit("2 km")
    assert pynbody.units.get_item_with_unit(ar2d, 1).units == "km"
    npt.assert_allclose(pynbody.units.get_item_with_unit(ar2d, 1), [4., 5., 6.])



import matplotlib.pylab as plt

import pynbody
import pynbody.plot.sph as sph

# load the snapshot and set to physical units
s = pynbody.load('testdata/g15784.lr.01024.gz')
s.physical_units()

# load the halos
h = s.halos()

# center on the largest halo and align the disk
pynbody.analysis.angmom.sideon(h[1])

# create the subplots
f, axs = plt.subplots(1,2,figsize=(14,6))

#create a simple slice showing the gas temperature, with velocity vectors overlaid
sph.velocity_image(h[1].g, vector_color="cyan", qty="temp",width=50,cmap="YlOrRd",
                   denoise=True,approximate_fast=False, subplot=axs[0], show_cbar = False)

#you can also make a stream visualization instead of a quiver plot
pynbody.analysis.angmom.faceon(h[1])
s['pos'].convert_units('Mpc')
sph.velocity_image(s.g, width='3 Mpc', cmap = "Greys_r", mode='stream', units='Msol kpc^-2',
                   density = 2.0, vector_resolution=100, vmin=1e-1,subplot=axs[1],
                   show_cbar=False, vector_color='black')



import numpy as np
import pytest
from scipy.io import FortranFile as FF

import pynbody
from pynbody.halo.adaptahop import (
    AdaptaHOPCatalogue,
    BaseAdaptaHOPCatalogue,
    NewAdaptaHOPCatalogue,
)


# Note: we do not use a module-wide fixture here to prevent caching of units
@pytest.fixture
def f():
    yield pynbody.load("testdata/output_00080")


@pytest.fixture
def halos(f):
    yield AdaptaHOPCatalogue(f)


def test_load_adaptahop_catalogue(halos):
    assert len(halos) == halos._headers["nhalos"] + halos._headers["nsubs"]


@pytest.mark.parametrize(
    ("path", "nhalos", "halo1_len", "halo2_len", "halo1_iord", "halo2_iord"),
    (("testdata/output_00080", 170, 235, 1201,
      [   48,  7468, 33923],
      [     91,    1203,    2703,    4151,    5539,    6907,    8535,   10227,   11355,
       12739,   14303,   15547,  992896,  993104  ,  993284,  993464,  993608,  993784,
      993932,  994152,  994288,  994452,  994596  ,  994720,  994896,  995060,  995216,
      995352,  995512,  995620,  995772,  995924  ,  996068,  996208,  996336,  996512,
      996624,  996784,  996908,  997080,  997224  ,  997408,  997552,  997692,  997852,
      998008,  998152,  998296,  998428,  998556  ,  998692,  998880,  999036,  999152,
      999300,  999412,  999556,  999700,  999912  , 1000056, 1000200, 1000356, 1000508,
     1000668, 1000796, 1000924, 1001084, 1001264  , 1001420, 1001584, 1001712, 1001892,
     1002064, 1002212, 1002392, 1002552, 1002700  , 1002876, 1003048, 1003204, 1003344,
     1003548, 1003700, 1003828, 1004020, 1004204  , 1004360, 1004536, 1004676, 1004824,
     1005032, 1005188, 1005308, 1005420, 1005560  , 1005716, 1005864, 1006064, 1006188,
     1006308, 1006436, 1006580, 1006712, 1006880  , 1007060, 1007160, 1007308, 1007460,
     1007620, 1007736, 1007872, 1008060, 1008160  , 1008276, 1008404, 1008548, 1008688,
     1010244, 1012248, 1031365, 1033027,]
      ),
     ("testdata/new_adaptahop_output_00080", 2, 22, 26,
      [852173],
      [762435,  21188,  29809])
     )
)
@pytest.mark.parametrize('load_all', [True, False])
def test_load_halo(path, nhalos, halo1_len, halo2_len, halo1_iord, halo2_iord, load_all):
    f = pynbody.load(path)
    halos = f.halos()
    assert isinstance(halos, BaseAdaptaHOPCatalogue)

    if load_all:
        halos.load_all()

    assert len(halos) == nhalos

    assert len(halos[2]) == halo2_len
    assert (halos[2].dm['iord'][::10] == halo2_iord).all()

    assert len(halos[1]) == halo1_len
    assert (halos[1].dm['iord'][::100] == halo1_iord).all()


def test_properties_are_simarrays(f, halos):
    halo = halos[1]

    # Test these properties exist and have the right dimensions
    # NOTE: we don't check that the units are the same, only
    #       that we can convert to the units below.
    properties = [
        ("m", "Msol"),
        ("pos_x", "kpc"),
        ("pos_y", "kpc"),
        ("pos_z", "kpc"),
        ("vel_x", "km s**-1"),
        ("vel_y", "km s**-1"),
        ("vel_z", "km s**-1"),
        ("angular_momentum_x", "Msol km s**-1 kpc"),
        ("angular_momentum_y", "Msol km s**-1 kpc"),
        ("angular_momentum_z", "Msol km s**-1 kpc"),
        ("max_distance", "kpc"),
        ("shape_a", "kpc"),
        ("shape_b", "kpc"),
        ("shape_c", "kpc"),
        ("kinetic_energy", "km**2 Msol s**-2"),
        ("potential_energy", "km**2 Msol s**-2"),
        ("total_energy", "km**2 Msol s**-2"),
        ("virial_radius", "kpc"),
        ("virial_mass", "Msol"),
        ("virial_temperature", "K"),
        ("virial_velocity", "km s**-1"),
        ("nfw_rho0", "Msol kpc**-3"),
        ("nfw_R_c", "kpc"),
        ("pos", "kpc"),
        ("vel", "km s**-1"),
    ]

    for prop, unit in properties:
        v = halo.properties[prop]
        assert isinstance(v, pynbody.array.SimArray)
        assert v.sim is f
        # Verify that they are convertible
        v.in_units(unit)


def test_physical_conversion_from_halo(f, halos):
    halo1 = halos[1]
    halo2 = halos[2]

    # Make sure the conversion is propagated to the parent
    fields = (
        halo1.properties["m"],
        halo1["mass"],
        f.dm["mass"],
        halo2.properties["m"]
    )
    for field in fields:
        assert field.units != "Msol"
    halo1.physical_units()
    for field in fields:
        assert field.units == "Msol"

    # Get another halo and make sure it is also in physical units
    halo3 = halos[3]
    assert halo3.properties["m"].units == "Msol"


def test_physical_conversion_from_halo_catalogue(f, halos):
    assert f.dm["mass"].units != "Msol"
    halos.physical_units()
    assert f.dm["mass"].units == "Msol"
    assert halos[1].properties["m"].units == "Msol"
    assert halos[1].dm["mass"].units == "Msol"


def test_physical_conversion_from_snapshot(f):
    # Convert then load
    assert f.dm["mass"].units != "Msol"
    f.physical_units()
    halos = f.halos()

    assert f.dm["mass"].units == "Msol"
    assert halos[1].properties["m"].units == "Msol"
    assert halos[1].dm["mass"].units == "Msol"

    # Load then convert
    f = pynbody.load("testdata/output_00080")
    assert f.dm["mass"].units != "Msol"

    halos = f.halos()
    f.physical_units()

    assert f.dm["mass"].units == "Msol"
    assert halos[1].properties["m"].units == "Msol"
    assert halos[1].dm["mass"].units == "Msol"

def test_get_group(f, halos):
    group_array = halos.get_group_array(family='dm')
    iord = f.dm["iord"]

    for halo_id in range(1, len(halos) + 1):
        mask = group_array == halo_id

        # Check that the indices
        # - read from halo (halos[halo_id]['iord'])
        # - obtained from get_group_array masking
        # are the same (in term of sets)
        iord_1 = np.sort(iord[mask])
        iord_2 = np.sort(halos[halo_id].dm["iord"])

        np.testing.assert_equal(iord_1, iord_2)


def test_halo_particle_ids(halos):
    halos.load_all()
    with FF(halos._fname, mode="r") as f:
        for halo_id in range(1, len(halos) + 1):
            # Manually read the particle ids and make sure pynbody is reading them as it should
            f._fp.seek(halos[halo_id].properties["file_offset"])

            f.read_ints()  # number of particles
            expected_members = f.read_ints("i")  # halo members

            np.testing.assert_equal(
                expected_members, halos[halo_id].dm["iord"]
            )


@pytest.mark.parametrize(
    ("fname", "Halo_T", "ans"),
    (
        (
            "testdata/output_00080/Halos/tree_bricks080",
            AdaptaHOPCatalogue,
            dict(_longint=False, _read_contamination=False),
        ),
        (
            "testdata/new_adaptahop_output_00080/Halos/tree_bricks080",
            NewAdaptaHOPCatalogue,
            dict(_longint=False, _read_contamination=True),
        ),
        (
            "testdata/EDGE_adaptahop_output/tree_bricks047_contam",
            NewAdaptaHOPCatalogue,
            dict(_longint=True, _read_contamination=True),
        ),
        (
            "testdata/EDGE_adaptahop_output/tree_bricks047_nocontam",
            NewAdaptaHOPCatalogue,
            dict(_longint=True, _read_contamination=False),
        ),
    ),
)
def test_longint_contamination_autodetection(f, fname, Halo_T, ans):
    # Note: we hack things a little bit here because
    # we just want to make sure the longint/contamination
    # flags are properly detected.

    halos = Halo_T(f, fname=fname)
    assert halos._longint == ans["_longint"]
    assert halos._read_contamination == ans["_read_contamination"]

def test_halo_iteration(halos):
    h = list(halos)

    assert len(h) == len(halos)
    assert h[0] is halos[1]
    assert h[-1] is halos[len(halos)]



import numpy as np
import numpy.testing as npt
import pytest

import pynbody


def test_a_to_t():
    """Test scalefactor -> time conversion for accuracy. See also issue #479"""
    f = pynbody.new() # for default cosmology
    ipoints = pynbody.analysis.cosmology._interp_points
    interp_a = np.linspace(0.005, 1.0, ipoints+1) # enough values to force interpolation rather than direct calculation
    interp_z = 1./interp_a - 1.
    with pytest.warns(RuntimeWarning, match=r"Assuming default value for property.*"):
        interp_t = pynbody.analysis.cosmology.age(f, z=interp_z)
    direct_aform = interp_a[::100]
    z = 1./direct_aform-1.
    direct_tform = pynbody.analysis.cosmology.age(f,z)
    npt.assert_almost_equal(direct_tform, interp_t[::100], decimal=4)

@pytest.mark.filterwarnings("ignore:invalid value encountered in log:RuntimeWarning")
def test_aform_saturation():
    """Test that NaN is returned when tform cannot be calculated from aform"""
    ipoints = pynbody.analysis.cosmology._interp_points
    f = pynbody.new(ipoints + 1)
    f['aform'] = np.linspace(0.0,1.1,ipoints+1)-0.05
    with pytest.warns(RuntimeWarning, match=r"Assuming default value for property.*"):
        tf = f['tform'][::100]
    assert tf[0]!=tf[0] # nan outside range
    assert tf[-1]!=tf[-1] # nan outside range
    assert (tf[1:-1]==tf[1:-1]).all() # no nans inside range



import glob
import os
import pickle
import sys

import numpy as np

import pynbody
import pynbody.analysis.profile as profile
import pynbody.filt as filt
import pynbody.plot as pp
import pynbody.plot.sph
import pynbody.units as units

simname = sys.argv[1]   # use first argument as simulation name to analyze
pp.plt.ion()            # plot in interactive mode to show all plots as
                        #     they are made

s = pynbody.load(simname) # load file into pynbody
h = s.halos()             # find and load halos (using AHF if in path)
diskf = filt.Disc('40 kpc','2 kpc')  # filter for particles in disk
notdiskf = filt.Not(filt.Disc('40 kpc','3 kpc')) # particles outside of disk
i=1
if (len(sys.argv) > 2):
    # if there's a second argument, it can be a photogenic file
    # and analysis will follow the halo that contains those particles
    photiords = np.genfromtxt(sys.argv[2],dtype='i8')
    frac = float(len(np.where(np.in1d(photiords,h[i]['iord']))[0]))/len(photiords)
    print('i: %d frac: %.2f'%(i,frac))
    while(((frac) < 0.5) & (i<100)):
        i=i+1
        frac = float(len(np.where(np.in1d(photiords,h[i]['iord']))[0]))/len(photiords)
        print('i: %d frac: %.2f'%(i,frac))
else:
    # otherwise follow largest halo with at least 2 stars
    while len(h[i].star) <2: i=i+1

if (i==100): sys.exit()
pynbody.analysis.angmom.faceon(h[i])  # align halo faceon
s.physical_units()   # change to physical units
